from .inline_cfg import forward_value, merge_subscope, inline_expr, get_forward_value
from ..ir_utils import ir, IrRewriter, add_to_list, IrVisitor, IrExprVisitor, IrExprRewriter
from pygears.typing import cast
from copy import copy
from ..ast.call import const_func_args
from pygears.util.utils import quiter


class VariableFinder(IrExprVisitor):
    def __init__(self):
        self.variables = set()

    def visit_AssignValue(self, node: ir.AssignValue):
        self.visit(node.target)
        self.visit(node.val)

    def visit_Name(self, node):
        if node.ctx == 'load':
            self.variables.add(node.name)


class RegisterBlockDetect(IrVisitor):
    def __init__(self, ctx):
        super().__init__()
        self.visited = set()
        self.ctx = ctx
        self.registers = {}
        # self.reg_scopes = {r: [None, None] for r in registers}

    def visit(self, node):
        super().visit(node)
        self.visited.add(id(node))

        for r, spec in self.registers.items():
            if spec['source'] is node:
                spec['source_scope'] = [s for s in self.scopes if isinstance(s, ir.LoopBlock)]

    def Branch(self, block: ir.Branch):
        self.ctx.reaching[id(block.test)] = self.ctx.reaching[id(block)]
        super().Branch(block)

    def LoopBlock(self, block: ir.LoopBlock):
        # Check if any of the variables are conditionaly changed (i.e.
        # under some if statement) in the loop. This means that their value
        # needs to be registered.

        stmt_id = self.parent.stmts.index(block)
        if stmt_id == 0:
            prev_stmt = self.parent
        else:
            prev_stmt = self.parent.stmts[stmt_id - 1]

        all_in_start = self.ctx.reaching[id(prev_stmt)]['out']
        all_in_end = self.ctx.reaching[id(block.stmts[-1])]['out']

        changed = set(name for name, _ in (all_in_end - all_in_start))

        loop_regs = set(name for name, _ in (all_in_end & all_in_start)
                        if name in changed)

        for name in loop_regs:
            reg_stat = {
                'target': block.stmts[-1],
                'target_scope': [s for s in self.scopes if isinstance(s, ir.LoopBlock)] + [block],
            }

            reg_stat['source'] = reg_stat['target']
            reg_stat['source_scope'] = reg_stat['target_scope']

            self.registers[name] = reg_stat

        self.ctx.reaching[id(block.test)] = {'in': self.ctx.reaching[id(block.stmts[-1])]['out']}
        super().LoopBlock(block)

    def Statement(self, expr):
        if id(expr) not in self.ctx.reaching:
            return

        if all(id(d[1].value) in self.visited for d in self.ctx.reaching[id(expr)].get('in', [])):
            return

        v = VariableFinder()
        v.visit(expr)

        if not v.variables:
            return

        for name, n in self.ctx.reaching[id(expr)]['in']:
            if name not in v.variables:
                continue

            if id(n.value) in self.visited:
                continue

            if name in self.registers:
                continue

            self.registers[name] = {
                'target': expr,
                'source': n.value,
                'target_scope': [s for s in self.scopes if isinstance(s, ir.LoopBlock)],
            }

    AssignValue = Statement
    Expr = Statement


class LoopBlockDetect(IrVisitor):
    def __init__(self):
        super().__init__()
        self.loops = []

    def LoopBlock(self, block: ir.LoopBlock):
        self.loops.append(block)
        self.BaseBlock(block)

    def Await(self, stmt: ir.Await):
        for s in self.scopes:
            if isinstance(s, ir.LoopBlock):
                s.blocking = True


def exhaust(func, args):
    try:
        return list(func(*args))
    except TypeError:
        raise SyntaxError(f'Generator "{func.__name__}" compile time invocation failed')


def ir_merge_subscope(block, block_scope_map):
    subscopes = [block_scope_map[b] for b in block.branches] + [block_scope_map[block]]
    tests = [b.test for b in block.branches] + [ir.res_true]

    return merge_subscope(subscopes, tests)


class Inline(IrRewriter):
    def __init__(self,
                 ctx,
                 scope_map=None,
                 scope_updater=forward_value,
                 scope_merger=ir_merge_subscope):
        super().__init__()
        self.scope_map = {} if scope_map is None else scope_map
        self.block_scope_map = {}
        self.ctx = ctx
        self.scope_updater = scope_updater
        self.scope_merger = scope_merger

    def enter_FuncBlock(self, block: ir.FuncBlock):
        for a in block.value.args:
            self.scope_map[a] = self.ctx.ref(a)

    def enter_Statement(self, block):
        self.block_scope_map[block] = copy(self.scope_map)

    def exit_HDLBlock(self, block: ir.HDLBlock):
        outscope = self.scope_merger(block, self.block_scope_map)
        self.scope_map.update(outscope)

    def handle_expr(self, expr):
        return inline_expr(expr, self.scope_map, self.ctx)

    def enter_FuncReturn(self, node: ir.FuncReturn):
        node.expr = self.handle_expr(node.expr)

    def enter_Branch(self, node: ir.Branch):
        self.scope_map = self.block_scope_map[self.parent].copy()
        # node.test = self.handle_expr(node.test)

    def exit_Branch(self, node: ir.Branch):
        self.block_scope_map[node] = self.scope_map.copy()

    def Expr(self, expr):
        return inline_expr(expr, self.scope_map, self.ctx)

    def AssignValue(self, node: ir.AssignValue):
        if isinstance(node.val, ir.GenNext):
            return self.AssignGenNext(node)

        rw_node = type(node)(val=self.handle_expr(node.val), target=self.visit(node.target))

        val = rw_node.val

        if isinstance(val, ir.ConcatExpr):
            val = ir.ConcatExpr(
                operands=[op.expr if isinstance(op, ir.Await) else op for op in val.operands])

        self.scope_updater(rw_node.target, val, self.scope_map)

        return rw_node


def forward_nonreg_value(target, val, scope):
    if isinstance(target, ir.Name):
        if not target.obj.reg or isinstance(val, ir.ResExpr):
            scope[target.name] = val
        else:
            scope[target.name] = ir.Name(target.name, target.obj, ctx='load')

        return True
    elif isinstance(target, ir.ConcatExpr):
        for i, t in enumerate(target.operands):
            forward_nonreg_value(t, ir.SubscriptExpr(val, ir.ResExpr(i)), scope)
    elif isinstance(target, ir.SubscriptExpr):
        if (isinstance(target.index, ir.ResExpr) and isinstance(val, ir.ResExpr)):
            base_val = get_forward_value(target.val, scope)

            if isinstance(base_val, ir.ResExpr):
                base_val.val[target.index.val] = cast(val.val, base_val.dtype[target.index.val])
                return True

        if isinstance(target.index, ir.ResExpr) and target.ctx == 'store':
            scope[f'{target.val.name}[{int(target.index.val)}]'] = val
            return True

        scope[target.val.name] = target.val


class Inliner(IrExprRewriter):
    def __init__(self, scope, ctx, missing_ok=False, reg_inits=None):
        self.scope_map = scope
        self.ctx = ctx
        self.missing_ok = missing_ok

        if reg_inits is None:
            reg_inits = {}
        self.reg_inits = reg_inits

    def visit_SubscriptExpr(self, node):
        if node.ctx == 'load':
            return super().visit_SubscriptExpr(node)
        else:
            return ir.SubscriptExpr(node.val, self.visit(node.index), ctx=node.ctx)

    def visit_Name(self, irnode):
        # If this name is target of the assignment, we have nothing to do
        if (irnode.ctx != 'load'):
            return None

        if (irnode.name not in self.scope_map):
            if self.missing_ok:
                return irnode

        val = self.scope_map[irnode.name]

        vv = VariableFinder()
        vv.visit(val)

        if any(self.ctx.scope[v].reg for v in vv.variables):
            return irnode

        if any(k.startswith(f'{irnode.name}[') for k in self.scope_map):
            elems = [ir.SubscriptExpr(irnode, ir.ResExpr(i)) for i in range(len(irnode.dtype))]
            for n in self.scope_map:
                if not n.startswith(f'{irnode.name}['):
                    continue

                index = int(n[n.index('[')+1:-1])
                elems[index] = self.scope_map[n]

            val = ir.ConcatExpr(elems)

        if isinstance(val, ir.Name) and val.obj.reg and val.name in self.reg_inits:
            return val.obj.val

        # TODO: What's with unknown?
        if isinstance(val, ir.ResExpr) and getattr(val.val, 'unknown', False):
            return irnode

        return val


def inline_expr(irnode, scope, ctx, reg_inits=None):
    new_node = Inliner(scope, ctx, reg_inits=reg_inits).visit(irnode)
    if new_node is None:
        return irnode

    return new_node


class LoopUnfolder(Inline):
    def __init__(self, ctx, scope_map=None):
        self.generators = {}
        super().__init__(ctx, scope_updater=forward_nonreg_value)

        for r in ctx.regs:
            self.scope_map[r] = ctx.ref(r)

    def handle_expr(self, expr):
        return inline_expr(expr, self.scope_map, self.ctx)

    def enter_RegReset(self, node):
        name = node.value.target.name
        if not self.ctx.scope[name].reg:
            self.scope_map[name] = node.value.target.val

    # def enter_HDLBlock(self, node):
    #     breakpoint()
    #     print("here")

    def LoopBlock(self, block: ir.LoopBlock):
        rw_block = type(block)(blocking=block.blocking)
        self.enter(rw_block)
        self.enter_scope(rw_block)

        last = ir.res_false
        while last == ir.res_false:
            for stmt in block.stmts:
                add_to_list(rw_block.stmts, self.visit(stmt))

            if rw_block.blocking:
                last = ir.res_true
                rw_block.test = self.visit(block.test)
            else:
                last = self.generators[block.test.val]['last']

        self.exit_scope()
        self.exit(rw_block)

        if rw_block.blocking:
            return rw_block
        else:
            return rw_block.stmts

    def ExprStatement(self, node):
        if not isinstance(node.expr, (ir.GenAck, ir.GenInit)):
            return super().ExprStatement(node)

        if isinstance(node.expr, ir.GenAck):
            if node.expr.val in self.generators:
                return None

        if isinstance(node.expr, ir.GenInit):
            gen_id = node.expr.val
            func_call = gen_id.obj.func

            if not const_func_args(func_call.args.values(), func_call.kwds):
                raise SyntaxError(f'Only generator calls with constant arguments are supported')

            args = tuple(node.val for node in func_call.args.values())

            vals = exhaust(func_call.func, args)

            self.generators[gen_id.name] = {'vals': quiter(vals)}

            return None

        return node

    def AssignGenNext(self, node: ir.AssignValue):
        gen_id = node.val.val
        if gen_id not in self.generators:
            breakpoint()

        next_val, last = next(self.generators[gen_id]['vals'])

        # self.forwarded[node.target.name] = ir.ResExpr(next_val)
        self.generators[gen_id]['last'] = ir.ResExpr(last)

        self.scope_updater(node.target, ir.ResExpr(next_val), self.scope_map)

        return None

    def AssignValue(self, node: ir.AssignValue):
        if isinstance(node.val, ir.GenNext):
            return self.AssignGenNext(node)
        else:
            return super().AssignValue(node)


def match_scopes(spec):
    for target_scope in reversed(spec['target_scope']):
        for source_scope in reversed(spec['source_scope']):
            if target_scope is source_scope:
                return target_scope
    else:
        # TODO: Is this possible?
        breakpoint()


def detect_loops(modblock, ctx):
    v = LoopBlockDetect()
    v.visit(modblock)

    ctx.unfoldable = any(not l.blocking for l in v.loops)

    # Discern which variables changed in the loop need to be registers
    v = RegisterBlockDetect(ctx)
    v.visit(modblock)

    reg_infer_dif = ctx.registers.symmetric_difference(v.registers)
    if reg_infer_dif:
        raise SyntaxError(f'Variable(s) "{reg_infer_dif}" used in different contexts in code. '
                          f'Compiler cannot determine whether to infer registers. Please rewrite'
                          f' the code so that usage of these variable(s) is clearer.')

    for r, spec in v.registers.items():
        scope = match_scopes(spec)
        ctx.scope[r].reg = scope.blocking


def loop_unfold(modblock, ctx):
    if ctx.unfoldable:
        v = LoopUnfolder(ctx)
        modblock = v.visit(modblock)

    return modblock



import ast
from . import Context, ir, node_visitor, visit_ast, visit_block
from pygears.lib.rng import qrange
from pygears.lib.union import select
from .stmt import assign_targets, extend_stmts
from .async_stmts import AsyncForContext, merge_cond_alias_map
from .inline import call_gear
from .generators import parse_generator_expression, is_intf_list


@node_visitor(ast.If)
def _(node: ast.If, ctx: Context):
    test_expr = visit_ast(node.test, ctx)

    ir_node = ir.HDLBlock()
    if test_expr != ir.res_false:
        branch = ir_node.add_branch(ir.Branch(test=test_expr))
        visit_block(branch, node.body, ctx)

    if test_expr != ir.res_true and hasattr(node, 'orelse') and node.orelse:
        orelse = ir.BaseBlock()
        visit_block(orelse, node.orelse, ctx)
        if orelse.stmts:
            if len(orelse.stmts) == 1 and isinstance(orelse.stmts[0], ir.HDLBlock):
                for b in orelse.stmts[0].branches:
                    ir_node.add_branch(b)
            else:
                branch = ir_node.add_branch(ir.Branch(stmts=orelse.stmts))
                ctx.closure_alias_maps[branch] = ctx.closure_alias_maps[orelse]

    if len(ir_node.branches) == 1 and ir_node.branches[0].test == ir.res_true:
        ctx.alias_map.update(ctx.closure_alias_maps[ir_node.branches[0]])
        return ir_node.branches[0].stmts

    for b in ir_node.branches:
        merge_cond_alias_map(b, ctx)

    if not ir_node.branches:
        return []

    return ir_node


@node_visitor(ast.While)
def _(node: ast.While, ctx: Context):
    test = visit_ast(node.test, ctx)
    if test == ir.res_false:
        return None

    ir_node = ir.HDLBlock()

    ir_node = ir.LoopBlock()
    loop = visit_block(ir_node, node.body, ctx)

    ctx.closures.append(ir_node)
    ir_node.test = visit_ast(node.test, ctx)
    ctx.closures.pop()

    merge_cond_alias_map(loop, ctx)

    maybe_loop = ir.HDLBlock([ir.Branch(test=test, stmts=[loop])])

    return maybe_loop


def intf_loop(node, intfs, targets, ctx: Context, enumerated):
    rng_intf, stmts = call_gear(qrange, [ir.ResExpr(len(intfs))], {}, ctx)
    ctx.ir_parent_block.stmts.extend(stmts)

    with AsyncForContext(rng_intf, ctx) as stmts:
        rng_iter = ir.SubscriptExpr(ir.Component(rng_intf.obj, 'data'), ir.ResExpr(0))
        select_intf, call_stmts = call_gear(select, args=[rng_iter] + intfs, kwds={}, ctx=ctx)
        ctx.ir_parent_block.stmts.extend(call_stmts)

        if enumerated:
            intf_var_name = targets.operands[1].name
        else:
            intf_var_name = targets.name

        ctx.local_namespace[intf_var_name] = select_intf

        if enumerated:
            extend_stmts(
                ctx.ir_parent_block.stmts,
                assign_targets(ctx, targets.operands[0],
                               ir.SubscriptExpr(ir.Component(rng_intf.obj, 'data'), ir.ResExpr(0))))

        for stmt in node.body:
            res_stmt = visit_ast(stmt, ctx)
            extend_stmts(ctx.ir_parent_block.stmts, res_stmt)

        return stmts


@node_visitor(ast.For)
def _(node: ast.For, ctx: Context):
    out_intf_ref, targets, gen_name = parse_generator_expression(node, ctx)

    if is_intf_list(out_intf_ref):
        return intf_loop(node, out_intf_ref.operands, targets, ctx,
                         getattr(out_intf_ref, 'enumerated', False))

    block = ir.LoopBlock(stmts=[ir.AssignValue(targets, ir.GenNext(gen_name))],
                         test=ir.GenDone(gen_name))

    visit_block(block, node.body, ctx)

    block.stmts.append(ir.ExprStatement(ir.GenAck(gen_name)))

    merge_cond_alias_map(block, ctx)

    return [ir.ExprStatement(ir.GenInit(ctx.ref(gen_name))), block]



from setuptools import setup, find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from setuptools.command.egg_info import egg_info
import os


def setup_home():
    os.makedirs(os.path.expanduser('~/.pygears'), exist_ok=True)
    os.makedirs(os.path.expanduser('~/.pygears/svlib'), exist_ok=True)


class PostEggInfoCommand(egg_info):
    def run(self):
        egg_info.run(self)
        setup_home()


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        setup_home()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        setup_home()


this_directory = os.path.abspath(os.path.dirname(__file__))


def readme():
    with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
        return f.read()


# TODO: Make it work for new jinja versions

setup(
    name='pygears',
    version='0.3.0',
    description='Framework for functional hardware design approach',
    long_description=readme(),
    url='https://www.pygears.org',
    # download_url = '',
    author='Bogdan Vukobratovic',
    author_email='bogdan.vukobratovic@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.6.0',
    package_data={'': ['*.j2', '*.v', '*.sv', '*.svt', '*.vt']},
    include_package_data=True,
    keywords='functional hardware design Python simulator HDL ASIC FPGA Gears',
    install_requires=[
        'jinja2', 'dataclasses;python_version<"3.7"', 'pyvcd', 'attrs', 'stopit'
    ],
    setup_requires=[
        'jinja2', 'dataclasses;python_version<"3.7"', 'pyvcd', 'attrs', 'stopit'
    ],
    packages=find_packages(exclude=['examples*', 'docs']),
    extras_require={'pytest': ['pytest', 'pytest-xdist', 'filelock']},
    entry_points={
        'console_scripts': [
            'pygears = pygears.entry:main',
        ],
        'pytest11': ['pytest_pygears = pygears.util.pytest_pygears']
    },
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
        'egg_info': PostEggInfoCommand,
    },
)



import ast as opc
import inspect
import attr
import typing
import textwrap
from dataclasses import dataclass, field

from pygears.typing.base import TypingMeta, GenericMeta, class_and_instance_method, is_type
from functools import reduce
from pygears import Intf
from pygears.core.port import InPort, OutPort
from pygears.core.gear import InSig, OutSig
from pygears.typing import (Bool, Integer, Queue, Tuple, Uint, is_type, typeof, Array, Union, Unit,
                            cast, code, Nothing)
# from .ast.utils import get_property_type
import operator

BOOLEAN_OPERATORS = {opc.BitOr, opc.BitAnd, opc.BitXor, opc.Invert, opc.Not, opc.And, opc.Or}
BIN_OPERATORS = [opc.Eq, opc.Gt, opc.GtE, opc.Lt, opc.LtE, opc.NotEq, opc.And, opc.Or]

COMMUTATIVE_BIN_OPERATORS = [
    opc.Add, opc.Mult, opc.BitOr, opc.BitAnd, opc.BitXor, opc.And, opc.Or, opc.Eq
]

ARITH_BIN_OPERATORS = [
    opc.Add, opc.Sub, opc.Mult, opc.Div, opc.Mod, opc.Pow, opc.LShift, opc.RShift, opc.BitOr,
    opc.BitAnd, opc.BitXor, opc.Div
]

EXTENDABLE_OPERATORS = [
    opc.Add, opc.Sub, opc.Mult, opc.Div, opc.Mod, opc.Pow, opc.LShift, opc.RShift, opc.BitOr,
    opc.BitAnd, opc.BitXor, opc.Div, opc.Invert, opc.Not
]

OPMAP = {
    opc.Add: '+',
    opc.Sub: '-',
    opc.MatMult: '@',
    opc.Mult: '*',
    opc.Div: '/',
    opc.Mod: '%',
    opc.Pow: '**',
    opc.LShift: '<<',
    opc.RShift: '>>',
    opc.BitOr: '|',
    opc.BitAnd: '&',
    opc.BitXor: '^',
    opc.FloorDiv: '/',
    opc.Invert: '~',
    opc.Not: '!',
    opc.UAdd: '+',
    opc.USub: '-',
    opc.Eq: '==',
    opc.Gt: '>',
    opc.GtE: '>=',
    opc.Lt: '<',
    opc.LtE: '<=',
    opc.NotEq: '!=',
    opc.And: '&&',
    opc.Or: '||'
}

PYOPMAP = {
    opc.Add: operator.__add__,
    opc.And: operator.__and__,
    opc.BitAnd: operator.__and__,
    opc.BitOr: operator.__or__,
    opc.BitXor: operator.__xor__,
    opc.Div: operator.__truediv__,
    opc.Eq: operator.__eq__,
    opc.Gt: operator.__gt__,
    opc.GtE: operator.__ge__,
    opc.FloorDiv: operator.__floordiv__,
    opc.Lt: operator.__lt__,
    opc.LtE: operator.__le__,
    opc.LShift: operator.__lshift__,
    opc.MatMult: operator.__matmul__,
    opc.Mult: operator.__mul__,
    opc.NotEq: operator.__ne__,
    opc.Not: operator.__not__,
    opc.Or: operator.__or__,
    opc.RShift: operator.__rshift__,
    opc.Sub: operator.__sub__,
    opc.UAdd: operator.__pos__,
    opc.USub: operator.__neg__,
}

REDUCE_INITIAL = {
    opc.Add: 0,
    opc.And: True,
    opc.BitAnd: True,
    opc.BitOr: False,
    opc.Mult: 1,
    opc.Or: False,
}


def opex(op, *operands):
    try:
        if op in BIN_OPERATORS or op is opc.Not:
            res_type = Bool
        else:
            res_type = PYOPMAP[op](*(p.dtype for p in operands))

        return res_type(PYOPMAP[op](*(p.val for p in operands)))
    except:
        return PYOPMAP[op](*(p.val for p in operands))


def find_sub_dtype(val):
    if isinstance(val, (tuple, list)):
        sub = max(val, key=lambda x: getattr(x, 'dtype').width)
        return type(sub)

    return type(val)


def find_name(node):
    name = getattr(node, 'name', None)
    if name is not None:
        return name

    if hasattr(node, 'val'):
        return find_name(node.val)

    if hasattr(node, 'op'):
        return find_name(node.op)

    return None


def get_contextpr(node):
    if isinstance(node, ResExpr):
        return node.val

    # TODO: rethink this? This should be some sort of named constant expression?
    if isinstance(node, Name):
        obj = node.obj
        if isinstance(obj, Variable) and obj.val is not None and not obj.reg:
            return obj.val

    return None


class IntfTypeMeta(GenericMeta):
    iin = 0
    iout = 1

    @property
    def dtype(self):
        return self.args[0]

    @property
    def direction(self):
        return self.args[1]


class IntfType(tuple, metaclass=IntfTypeMeta):
    __parameters__ = ['dtype', 'direction']

    def pull_nb(self):
        pass

    def empty(self):
        pass

    def ack(self):
        pass


@attr.s(auto_attribs=True, kw_only=True)
class Expr:
    @property
    def dtype(self):
        pass


# Type aliases

PgType = typing.Union[TypingMeta, Unit, int]
OpType = typing.Union[Expr, PgType, str]

# Type definitions


class EmptyTypeMeta(GenericMeta):
    @property
    def dtype(self):
        if self.args:
            return self.args[0]
        else:
            return None


class EmptyType(metaclass=EmptyTypeMeta):
    __parameters__ = ['dtype']

    def __init__(self, v=None):
        if (v is not None) and (not isinstance(v, EmptyType)):
            raise TypeError

    def __hash__(self):
        if type(self).dtype is None:
            return hash(type(self))
        else:
            return hash((type(self), type(self).dtype))

    def __repr__(self):
        if type(self).dtype is None:
            return 'Empty'
        else:
            return f'{repr(type(self).dtype)}(Empty)'


class ResExpr(Expr):
    def __new__(cls, val):
        if isinstance(val, Expr):
            return val

        inst = super().__new__(cls)

        # TODO: Think about this automatic casting. For example when register
        # is inferred based on this value, it might be wrong. ex. "cnt = 0"
        if not is_type(type(val)) and isinstance(val, (int, float)):
            val = cast(val, Integer)

        inst.val = val

        return inst

    def __init__(self, val):
        super().__init__()

    def __eq__(self, other):
        if not isinstance(other, ResExpr):
            return False

        return self.val == other.val

    def __hash__(self):
        return hash((self.val, type(self)))

    def __repr__(self):
        return f'ResExpr({repr(self.val)})'

    def __str__(self):
        return str(self.val)

    @property
    def dtype(self):
        if isinstance(self.val, EmptyType):
            if type(self.val).dtype is None:
                return EmptyType
            else:
                return type(self.val).dtype

        # if is_type(type(self.val)):
        #     return type(self.val)

        if not is_type(type(self.val)) and isinstance(self.val, int):
            return type(Integer(self.val))

        # TODO: Remove this if unecessary
        if isinstance(self.val, Intf):
            return IntfType[self.val.dtype]

        return type(self.val)

        # return None


res_true = ResExpr(Bool(True))
res_false = ResExpr(Bool(False))


@dataclass
class TupleExpr(Expr):
    val: typing.Sequence

    def __getitem__(self, key):
        return self.val[key]


@dataclass
class Variable:
    name: str
    dtype: typing.Union[PgType, typing.Any] = None
    val: typing.Union[PgType, Expr] = None
    any_init: Bool = False
    reg: Bool = False

    def __post_init__(self):
        if self.dtype is None:
            if self.val is not None:
                self.dtype = self.val.dtype


@dataclass
class Interface(Expr):
    intf: typing.Union[InPort, OutPort, Expr]
    direction: str
    _name: str = None

    @property
    def name(self):
        if self._name:
            return self._name
        try:
            return self.intf.basename
        except AttributeError:
            return find_name(self.intf)

    def __str__(self):
        return self.name

    @property
    def dtype(self):
        return find_sub_dtype(self.intf)


@attr.s(auto_attribs=True)
class Name(Expr):
    name: str
    obj: Variable = None
    ctx: str = 'load'

    def __repr__(self):
        return f'Id({self.name})'

    def __str__(self):
        if self.ctx in ['load', 'store']:
            return self.name
        else:
            return f'{self.name}.{self.ctx}'

    @property
    def dtype(self):
        return self.obj.dtype


@attr.s(auto_attribs=True)
class Component(Expr):
    val: Expr
    field: str

    def __repr__(self):
        return f'{repr(self.val)}.{self.field}'

    def __str__(self):
        return f'{self.val}.{self.field}'

    @property
    def dtype(self):
        if self.field in ['ready', 'valid']:
            return Bool
        elif self.field == 'data':
            assert typeof(self.val.dtype, IntfType)
            return self.val.dtype.dtype


@dataclass
class ConcatExpr(Expr):
    def __repr__(self):
        return 'ConcatExpr(' + ', '.join([repr(v) for v in self.operands]) + ')'

    def __str__(self):
        return '(' + ', '.join([str(v) for v in self.operands]) + ')'

    def __init__(self, operands: typing.Sequence[Expr]):
        pass

    def __eq__(self, other):
        if not isinstance(other, BinOpExpr):
            return False

        return all(ops == opo for ops, opo in zip(self.operands, other.operands))

    def __new__(cls, operands: typing.Sequence[Expr]):
        if all(isinstance(v, ResExpr) for v in operands):
            if (all(is_type(v.dtype) for v in operands)
                    and not any(isinstance(v.val, EmptyType) for v in operands)):
                return ResExpr(Tuple[tuple(v.dtype for v in operands)](tuple(v.val
                                                                             for v in operands)))
            else:
                return ResExpr(tuple(v.val for v in operands))

        inst = super().__new__(cls)
        inst.operands = operands

        return inst

    @property
    def dtype(self):
        return Tuple[tuple(op.dtype for op in self.operands)]


class UnaryOpExpr(Expr):
    def __init__(self, operand, operator):
        pass

    def __repr__(self):
        return f'{OPMAP[self.operator]}({self.operand})'

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (self.operand == other.operand and self.operator == other.operator)

    def __new__(cls, operand, operator):
        if isinstance(operand, ResExpr):
            return ResExpr(opex(operator, operand))

        if operator == opc.Not and isinstance(operand, BinOpExpr):
            if operand.operator == opc.Eq:
                return BinOpExpr(operand.operands, opc.NotEq)

            if operand.operator == opc.NotEq:
                return BinOpExpr(operand.operands, opc.Eq)

            # De-Morgan laws
            if operand.operator in [opc.And, opc.Or]:
                operands = [UnaryOpExpr(op, opc.Not) for op in operand.operands]
                return BinOpExpr(operands, opc.Or if operand.operator == opc.And else opc.And)

        elif operator == opc.Not and isinstance(operand, ConditionalExpr):
            op1, op2 = operand.operands
            if typeof(op1.dtype, (Uint[1], Bool)) and typeof(op2.dtype, (Uint[1], Bool)):
                return ConditionalExpr([UnaryOpExpr(op1, opc.Not),
                                        UnaryOpExpr(op2, opc.Not)], operand.cond)

        elif operator == opc.Not and isinstance(operand,
                                                UnaryOpExpr) and operand.operator == opc.Not:
            return operand.operand

        inst = super().__new__(cls)
        inst.operand = operand
        inst.operator = operator
        return inst

    @property
    def dtype(self):
        if self.operator == opc.Not:
            return Uint[1]

        res_t = eval(f'{OPMAP[self.operator]} op', {'op': self.operand.dtype})

        if isinstance(res_t, bool):
            return Uint[1]

        return res_t


class CastExpr(Expr):
    def __init__(self, operand, cast_to):
        pass

    def __repr__(self):
        return f'CastTo({self.operand}, {self.cast_to})'

    def __str__(self):
        return f"({self.cast_to})'({self.operand})"

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (self.operand == other.operand and self.cast_to == other.cast_to)

    def __new__(cls, operand, cast_to):
        if isinstance(cast_to, ResExpr):
            cast_to = cast_to.val

        if cast_to == int:
            cast_to = Uint[operand.dtype.width]

        if operand.dtype == cast_to:
            return operand

        if isinstance(operand, ResExpr):
            return ResExpr(code(operand.val, cast_to))

        if isinstance(operand, ConcatExpr) and typeof(cast_to, (Array, Tuple, Queue, Union)):
            cast_ops = [
                CastExpr(op, cast_t) if op.dtype != cast_t else op
                for op, cast_t in zip(operand.operands, cast_to)
            ]
            operand = ConcatExpr(cast_ops)

        inst = super().__new__(cls)
        inst.operand = operand
        inst.cast_to = cast_to
        return inst

    @property
    def dtype(self):
        return self.cast_to


class SliceExpr(Expr):
    def __repr__(self):
        return f'({self.start if self.start else ""}:{self.stop if self.stop else ""}:{self.step if self.step else ""})'

    def __init__(self, start: OpType, stop: OpType, step: OpType):
        pass

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (self.start == other.start and self.stop == other.stop and self.step == other.step)

    def __new__(cls, start: OpType, stop: OpType, step: OpType):
        if isinstance(start, ResExpr) and isinstance(stop, ResExpr) and isinstance(step, ResExpr):
            return ResExpr(slice(start.val, stop.val, step.val))

        inst = super().__new__(cls)
        inst.start = start
        inst.stop = stop
        inst.step = step
        return inst


def identity_anihilation(op1, op2, operator):
    '''
    x && True => x
    x || False => x

    x && False => False
    x || True => True
    '''
    if operator == opc.And:
        if isinstance(op1, ResExpr):
            return op2 if op1.val else op1

    elif operator == opc.Or:
        if isinstance(op1, ResExpr):
            return op1 if op1.val else op2

    return None


def idempotence(op1, op2, operator):
    '''
    x && x => x
    x || x => x
    '''
    if operator in [opc.And, opc.Or, opc.BitAnd, opc.BitOr]:
        if op1 == op2:
            return op1

    return None


def complementation(op1, op2, operator):
    '''
    x && (~x) => False
    x || (~x) => True
    '''
    if isinstance(op1, UnaryOpExpr) and op1.operator == opc.Not:
        if op1.operand == op2:
            if operator is opc.And:
                return res_false
            elif operator is opc.Or:
                return res_true

    return None


def absorption(op1, op2, operator):
    '''
    x && (x || y) => x
    x || (x && y) => x

    x && ((~x) || y) => x && y
    x || ((~x) && y) => x || y
    '''
    if not isinstance(op2, BinOpExpr):
        return None

    op2_1, op2_2 = op2.operands
    if ((operator is opc.And and op2.operator is opc.Or)
            or (operator is opc.Or and op2.operator is opc.And)):
        if op1 == op2_1:
            return op2_1
        elif op1 == op2_2:
            return op2_2
        elif op1 == UnaryOpExpr(op2_1, opc.Not):
            return BinOpExpr([op1, op2_2], operator)
        elif op1 == UnaryOpExpr(op2_2, opc.Not):
            return BinOpExpr([op1, op2_1], operator)

    return None


def elimination(op1, op2, operator):
    '''
    (x && y) || (x && (~y)) => x
    (x || y) && (x || (~y)) => x
    '''
    if not (isinstance(op1, BinOpExpr) and isinstance(op2, BinOpExpr)):
        return None

    if op1.operator not in [opc.And, opc.Or] or op2.operator not in [opc.And, opc.Or]:
        return None

    if op1.operator != op2.operator or op1.operator == operator:
        return None

    op2_1, op2_2 = op2.operands
    for op1_1, op1_2 in [op1.operands, op1.operands[::-1]]:
        if op1_1 == op2_1 and op1_2 == UnaryOpExpr(op2_2, opc.Not):
            return op1_1
        elif op1_1 == op2_2 and op1_2 == UnaryOpExpr(op2_1, opc.Not):
            return op1_1

    return None


# def elimination_same(op1, op2, operator):
#     '''
#     x && (x && y) => x && y
#     x || (x || y) => x || y
#     x && ((~x) && y) => y
#     x || ((~x) || y) => True
#     '''
#     if not isinstance(op2, BinOpExpr) or operator != op2.operator:
#         return None

#     if op1.operator not in [opc.And, opc.Or] or op2.operator not in [opc.And, opc.Or]:
#         return None

#     if op1.operator != op2.operator or op1.operator == operator:
#         return None

#     op2_1, op2_2 = op2.operands
#     for op1_1, op1_2 in [op1.operands, op1.operands[::-1]]:
#         if op1_1 == op2_1 and op1_2 == UnaryOpExpr(op2_2, opc.Not):
#             return op1_1
#         elif op1_1 == op2_2 and op1_2 == UnaryOpExpr(op2_1, opc.Not):
#             return op1_1

#     return None


def booleq(op1, op2, operator):
    if operator not in [opc.Eq, opc.NotEq]:
        return

    if not (typeof(op1.dtype, (Bool, Uint[1])) and typeof(op2.dtype, (Bool, Uint[1]))):
        return

    if op2 == res_true:
        return op1 if operator == opc.Eq else UnaryOpExpr(op1, opc.Not)

    if op2 == res_false:
        return op1 if operator == opc.NotEq else UnaryOpExpr(op1, opc.Not)

    return None


def conditional_distribution(op1, op2, operator):
    if operator not in ARITH_BIN_OPERATORS:
        return None

    if (isinstance(op2, ConditionalExpr) and isinstance(op2.operands[0], ResExpr)
            and isinstance(op2.operands[1], ResExpr)):

        return ConditionalExpr([
            ResExpr(opex(operator, op1, op2.operands[0])),
            ResExpr(opex(operator, op1, op2.operands[1]))
        ], op2.cond)


bin_op_transforms = [
    identity_anihilation,
    idempotence,
    complementation,
    absorption,
    elimination,
    booleq,
    conditional_distribution,
]


# TODO: Should be allow BinOpExpr to have arbitrary number of boolean
# variables. This would allow for some simplfifications, ex. : (~x && y) && (x && ~y)
class BinOpExpr(Expr):
    def __repr__(self):
        try:
            return (f'{type(self).__name__}(operands={repr(self.operands)}, '
                    f'operator={self.operator.__name__})')
        except:
            return (f'{type(self).__name__}(operands={repr(self.operands)}, '
                    f'operator={self.operator})')

    def __str__(self):
        return f'({self.operands[0]} {OPMAP[self.operator]} {self.operands[1]})'

    def __init__(self, operands: typing.Tuple[OpType], operator):
        pass

    def __new__(cls, operands: typing.Tuple[OpType], operator):
        if len(operands) > 2:
            op1 = operands[0]
            op2 = BinOpExpr(operands[1:], operator)
        else:
            op1, op2 = operands

        if isinstance(op1, ResExpr) and isinstance(op2, ResExpr):
            return ResExpr(opex(operator, op1, op2))

        for t in bin_op_transforms:
            ret = t(op1, op2, operator)
            if ret is not None:
                return ret

        if operator in COMMUTATIVE_BIN_OPERATORS:
            for t in bin_op_transforms:
                ret = t(op2, op1, operator)
                if ret is not None:
                    return ret

        if operator in (opc.RShift, opc.LShift):
            if isinstance(op2, ResExpr) and op2.val == 0:
                return op1

        inst = super().__new__(cls)
        inst.operands = operands
        inst.operator = operator
        return inst

    def __eq__(self, other):
        if not isinstance(other, BinOpExpr):
            return False

        return ((self.operator == other.operator)
                and (all(ops == opo for ops, opo in zip(self.operands, other.operands))))

    @property
    def dtype(self):
        if self.operator in BIN_OPERATORS:
            return Uint[1]

        if (self.operator in (opc.LShift, opc.RShift)) and isinstance(self.operands[1], ResExpr):
            op2 = self.operands[1].val
        else:
            op2 = self.operands[1].dtype

        res_t = eval(f'op1 {OPMAP[self.operator]} op2', {'op1': self.operands[0].dtype, 'op2': op2})

        if isinstance(res_t, bool):
            return Uint[1]

        return res_t


class ArrayOpExpr(Expr):
    def __repr__(self):
        return f'{type(self).__name__}(val={repr(self.array)}, op={repr(self.operator)})'

    def __str__(self):
        return f'{OPMAP[self.operator]}({str(self.array)})'

    def __init__(self, array: Expr, operator):
        pass

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.operator == other.operator and self.array == other.array

    def __new__(cls, array: Expr, operator):
        if isinstance(array, ResExpr):
            return ResExpr(reduce(PYOPMAP[operator], array.val, REDUCE_INITIAL[operator]))

        inst = super().__new__(cls)
        inst.array = array
        inst.operator = operator

        return inst

    @property
    def dtype(self):
        return self.array.dtype[0]


class SubscriptExpr(Expr):
    def __repr__(self):
        return f'{type(self).__name__}(val={repr(self.val)}, index={repr(self.index)}, ctx={self.ctx})'

    def __str__(self):
        return f'{self.val}[{self.index}]'

    def __init__(self, val: Expr, index: Expr, ctx: str = 'load'):
        pass

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.val == other.val and self.index == other.index

    def __new__(cls, val: Expr, index: Expr, ctx: str = 'load'):
        const_index = get_contextpr(index)
        const_val = get_contextpr(val)
        if const_index is not None:
            if const_val is not None:
                return ResExpr(const_val[const_index])

            if isinstance(val, ConcatExpr):
                return val.operands[const_index]

        inst = super().__new__(cls)
        inst.val = val
        inst.index = index
        inst.ctx = ctx

        if inst.dtype.width == 0:
            return ResExpr(Unit())

        return inst

    @property
    def dtype(self):
        val_dtype = self.val.dtype
        if val_dtype is None:
            return None

        if isinstance(self.index, ResExpr):
            return self.val.dtype[self.index.val]

        # TODO: When is this usefull?
        return self.val.dtype[0]


class AttrExpr(Expr):
    def __str__(self):
        return f'{self.val}.{self.attr}'

    def __init__(self, val: Expr, attr: str):
        pass

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.val == other.val and self.attr == other.attr

    def __new__(cls, val: Expr, attr: str):
        const_val = get_contextpr(val)
        if const_val is not None:
            return ResExpr(getattr(const_val, attr))

        inst = super().__new__(cls)
        inst.val = val
        inst.attr = attr

        return inst

    @property
    def dtype(self):
        return getattr(self.val.dtype, self.attr, None)


class ConditionalExpr(Expr):
    def __repr__(self):
        return f'({self.cond} ? {self.operands[0]} : {self.operands[1]})'

    def __init__(self, operands: typing.Sequence[OpType], cond: Expr):
        pass

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return (self.cond == other.cond and self.operands[0] == other.operands[0]
                and self.operands[1] == other.operands[1])

    def __new__(cls, operands: typing.Sequence[OpType], cond: Expr):
        op1, op2 = operands

        const_cond = get_contextpr(cond)
        if const_cond is not None:
            return op1 if const_cond else op2

        if op1 == op2:
            return op1

        # TODO: Bool should be equivalent to Uint[1]
        if typeof(op1.dtype, (Uint[1], Bool)) and typeof(op2.dtype, (Uint[1], Bool)):
            if cond == op2:
                return BinOpExpr((cond, op1), opc.And)

            const_op1 = get_contextpr(op1)
            if const_op1 is not None:
                if const_op1:
                    return BinOpExpr((cond, op2), opc.Or)
                else:
                    return BinOpExpr((UnaryOpExpr(cond, opc.Not), op2), opc.And)

            const_op2 = get_contextpr(op2)
            if const_op2 is not None:
                if const_op2:
                    return BinOpExpr((UnaryOpExpr(cond, opc.Not), op1), opc.Or)
                else:
                    return BinOpExpr((cond, op1), opc.And)

        inst = super().__new__(cls)
        inst.operands = operands
        inst.cond = cond

        return inst

    @property
    def dtype(self):
        if self.operands[0].dtype.width > self.operands[1].dtype.width:
            return self.operands[0].dtype
        else:
            return self.operands[1].dtype

        # return max([op.dtype for op in self.operands])


@dataclass
class FunctionCall(Expr):
    name: str
    operands: typing.Tuple[OpType]
    keywords: typing.Dict[str, typing.Any] = None
    ret_dtype: PgType = None

    @property
    def dtype(self):
        return self.ret_dtype

    def __str__(self):
        ops = ', '.join([str(op) for op in self.operands])

        kwds = None
        if self.keywords is not None:
            kwds = ', '.join([f'{n}={v}' for n, v in self.keywords.items()])

        if not ops and not kwds:
            sig = ''
        elif not kwds:
            sig = ops
        else:
            sig = ', '.join([ops, kwds])

        return f'{self.name}({sig})'


@dataclass
class CallExpr(Expr):
    func: typing.Any
    args: typing.Tuple[OpType]
    kwds: typing.Dict[str, typing.Any] = field(default_factory=dict)
    params: typing.Dict = field(default_factory=dict)
    pass_eot = True

    @property
    def dtype(self):
        ret = self.params['return']
        if isinstance(ret, tuple):
            ret = Tuple[ret]

        if not self.pass_eot:
            return ret.data

        return ret


@dataclass
class Generator:
    name: str
    func: CallExpr

    @property
    def dtype(self):
        return self.func.dtype


@attr.s(auto_attribs=True)
class GenDone(Expr):
    val: Name

    @property
    def dtype(self):
        return Bool


@attr.s(auto_attribs=True)
class GenInit(Expr):
    val: Name

    @property
    def dtype(self):
        return self.val.dtype


@attr.s(auto_attribs=True)
class GenNext(Expr):
    val: Name

    @property
    def dtype(self):
        return self.val.dtype


@attr.s(auto_attribs=True)
class GenAck(Expr):
    val: Name

    @property
    def dtype(self):
        return self.val.dtype


# Statements


# @attr.s(auto_attribs=True)
@attr.s(auto_attribs=True, kw_only=True, eq=False)
class Statement:
    def __hash__(self):
        return id(self)


@attr.s(auto_attribs=True, eq=False)
class Await(Statement):
    expr: Expr

    def __str__(self):
        return f'(await {str(self.expr)})\n'


@attr.s(auto_attribs=True, eq=False)
class Jump(Statement):
    label: str
    where: typing.Any = None

    def __str__(self):
        if self.where is None:
            return f'->> {str(self.label)}\n'
        else:
            return f'->> {self.label} {self.where}\n'


@attr.s(auto_attribs=True, eq=False)
class ExprStatement(Statement):
    expr: Expr

    def __str__(self):
        return f'{self.expr}\n'


@attr.s(auto_attribs=True, eq=False)
class RegReset(Statement):
    target: Expr

    def __str__(self):
        return f'reset {self.target}\n'


@attr.s(auto_attribs=True, eq=False)
class Assert(Statement):
    test: Expr
    msg: str = None

    def __str__(self):
        return f'{self.test}, "{self.msg}"\n'


def extract_base_targets(target):
    if isinstance(target, ConcatExpr):
        for t in target.operands:
            yield from extract_base_targets(t)
    elif isinstance(target, Name) and isinstance(target.obj, Variable):
        yield target


def extract_partial_targets(target):
    if isinstance(target, SubscriptExpr):
        yield from extract_base_targets(target.val)


@attr.s(auto_attribs=True, eq=False)
class AssignValue(Statement):
    target: Union[str, Name]
    val: Union[str, int, Expr]
    dtype: Union[TypingMeta, None] = None

    def __attrs_post_init__(self):
        for t in extract_base_targets(self.target):
            t.ctx = 'store'

    @property
    def base_targets(self):
        return list(extract_base_targets(self.target))

    @property
    def partial_targets(self):
        return list(extract_partial_targets(self.target))

    def __str__(self):
        return f'{str(self.target)} <= {str(self.val)}\n'


@dataclass
class AssertValue(Statement):
    val: typing.Any


# Blocks


@attr.s(auto_attribs=True, kw_only=True, eq=False)
class BaseBlock(Statement):
    stmts: typing.List = attr.Factory(list)

    def __attrs_post_init__(self):
        if self.stmts is None:
            self.stmts = []

    def __str__(self):
        body = ''
        for s in self.stmts:
            body += str(s)

        return f'{{\n{textwrap.indent(body, "    ")}}}\n'


@attr.s(auto_attribs=True, kw_only=True, eq=False)
class Module(BaseBlock):
    states: typing.List = attr.Factory(list)
    funcs: typing.List = attr.Factory(list)


@attr.s(auto_attribs=True, eq=False)
class Branch(BaseBlock):
    test: Expr = res_true

    def __str__(self):
        if self.test == res_true and not self.stmts:
            return ''

        if self.test == res_true:
            header = ''
        else:
            header = f'if ({str(self.test)})'

        footer = ''

        body = ''
        for s in self.stmts:
            body += str(s)

        return f'{header}{{\n{textwrap.indent(body, "    ")}}}{footer}'


@attr.s(auto_attribs=True, eq=False)
class LoopBody(BaseBlock):
    state_id: int = None


@attr.s(auto_attribs=True, eq=False)
class HDLBlock(Statement):
    branches: typing.List[Branch] = attr.Factory(list)

    def __init__(self, branches=None):
        if branches is not None:
            branches = [b for b in branches if b.test != res_false]

        super().__init__(branches)

    def add_branch(self, branch: Branch = None):
        if branch is None:
            branch = Branch()

        self.branches.append(branch)
        return branch

    @property
    def has_else(self):
        return self.branches[-1].test == res_true

    def __str__(self):
        sblk = ''
        for i, b in enumerate(self.branches):
            if b.stmts:
                if i > 0:
                    sblk += ' else '

                sblk += str(b)

        sblk += '\n'

        return sblk


@attr.s(auto_attribs=True, eq=False)
class LoopBlock(BaseBlock):
    test: Expr = res_true
    blocking: Expr = False

    def __str__(self):
        return f'do ' + super().__str__() + f'while ({self.test})\n'


@attr.s(auto_attribs=True, eq=False)
class BaseBlockSink(Statement):
    def __str__(self):
        return f'BaseBlockSink'


@attr.s(auto_attribs=True, eq=False)
class ModuleSink(BaseBlockSink):
    def __str__(self):
        return f'ModuleSink'


@attr.s(auto_attribs=True, eq=False)
class BranchSink(BaseBlockSink):
    def __str__(self):
        return f'BranchSink'


@attr.s(auto_attribs=True, eq=False)
class HDLBlockSink(Statement):
    def __str__(self):
        return f'HDLBlockSink'


@attr.s(auto_attribs=True, eq=False)
class LoopBlockSink(BaseBlock):
    def __str__(self):
        return f'LoopBlockSink'


@attr.s(auto_attribs=True, eq=False)
class FuncBlock(Module):
    args: typing.List[Name]
    name: str
    ret_dtype: PgType
    funcs: typing.List = attr.Factory(list)

    def __str__(self):
        body = ''
        for s in self.stmts:
            body += str(s)

        args = [f'{name}: {val}' for name, val in self.args.items()]
        return f'{self.name}({", ".join(args)}) {{\n{textwrap.indent(body, "    ")}}}\n'


@attr.s(auto_attribs=True, eq=False)
class FuncReturn(Statement):
    func: FuncBlock
    expr: Expr

    def __str__(self):
        return f'return {self.expr}'



from pygears import alternative, gear, module, reg
from pygears.core.intf import IntfOperPlugin
from pygears.typing import Any, Bool, Integer, Integral, Number, Tuple
from pygears.typing import div as typing_div, is_type, Uint, typeof
from pygears.typing import code as type_code
from pygears.util.hof import oper_tree
from pygears import datagear
from pygears.hdl.util import HDLGearHierVisitor
from pygears.hdl import flow_visitor
from pygears.hdl.sv import SVGenPlugin
from pygears.hdl.v import VGenPlugin

# TODO: Think about operand promoting (casting). What if a: Int[32] and someone
# writes a == 0? Which type should the 0 be?

@datagear
def add(din: Tuple[{'a': Number, 'b': Number}]) -> b'din[0] + din[1]':
    return din[0] + din[1]


@alternative(add)
@gear(enablement=b'len(din) > 2')
def add_vararg(*din: Integer):
    return oper_tree(din, add)


@datagear
def and_(din: Tuple[Integral, Integral]) -> b'din[0] & din[1]':
    return din[0] & din[1]


@datagear
def fdiv(din: Tuple[{'a': Integer, 'b': Integer}]) -> b'din[0] // din[1]':
    return din[0] // din[1]


@datagear
def cat(din: Tuple[{'a': Uint, 'b': Uint}]) -> b'din[0] @ din[1]':
    return din[0] @ din[1]


@datagear
def div(
        din: Tuple[{
            'a': Number,
            'b': Number
        }], *, subprec) -> b'typing_div(din[0], din[1], subprec)':
    return typing_div(din[0], din[1], subprec)


@datagear
def eq(din: Tuple[{'a': Any, 'b': Any}]) -> Bool:
    return din[0] == din[1]


@datagear
def ge(din: Tuple[{'a': Number, 'b': Number}]) -> Bool:
    return din[0] >= din[1]


@datagear
def gt(din: Tuple[{'a': Number, 'b': Number}]) -> Bool:
    return din[0] > din[1]


@datagear
def invert(a) -> b'a':
    return ~a


@datagear
def le(din: Tuple[{'a': Number, 'b': Number}]) -> Bool:
    return din[0] <= din[1]


@datagear
def lt(din: Tuple[{'a': Number, 'b': Number}]) -> Bool:
    return din[0] < din[1]


@datagear
def mod(din: Tuple[Integer, Integer]) -> b'din[0] % din[1]':
    return din[0] % din[1]


@datagear
def mul(din: Tuple[{'a': Number, 'b': Number}]) -> b'din[0] * din[1]':
    return din[0] * din[1]


@datagear
def ne(din: Tuple[Any, Any]) -> Bool:
    return din[0] != din[1]


@datagear
def neg(a: Number) -> b'-a':
    return -a


@datagear
def or_(din: Tuple[Integral, Integral]) -> b'din[0] | din[1]':
    return din[0] | din[1]


@datagear
def sub(din: Tuple[{'a': Number, 'b': Number}]) -> b'din[0] - din[1]':
    return din[0] - din[1]


@datagear
def shl(din: Integral, *, shamt) -> b'din << shamt':
    return module().tout(din << shamt)


@datagear
def shr(din: Integral, *, shamt) -> b'din >> shamt':
    return module().tout(din >> shamt)


@datagear
def code(din, *, t) -> b't':
    return type_code(din, t)


@datagear
def xor(din: Tuple[Integral, Integral]) -> b'din[0] ^ din[1]':
    return din[0] ^ din[1]


def shr_or_code(x, y):
    if is_type(y):
        if typeof(y, Uint) and (not y.specified):
            y = Uint[x.dtype.width]

        return code(x, t=y)
    else:
        return shr(x, shamt=y)


class AddIntfOperPlugin(IntfOperPlugin):
    @classmethod
    def bind(cls):
        reg['gear/intf_oper/__add__'] = add
        reg['gear/intf_oper/__and__'] = and_
        reg['gear/intf_oper/__floordiv__'] = fdiv
        reg['gear/intf_oper/__truediv__'] = or_
        reg['gear/intf_oper/__eq__'] = eq
        reg['gear/intf_oper/__ge__'] = ge
        reg['gear/intf_oper/__gt__'] = gt
        reg['gear/intf_oper/__invert__'] = invert
        reg['gear/intf_oper/__le__'] = le
        reg['gear/intf_oper/__lt__'] = lt
        reg['gear/intf_oper/__lshift__'] = lambda x, y: shl(x, shamt=y)
        reg['gear/intf_oper/__matmul__'] = cat
        reg['gear/intf_oper/__mod__'] = mod
        reg['gear/intf_oper/__mul__'] = mul
        reg['gear/intf_oper/__ne__'] = ne
        reg['gear/intf_oper/__neg__'] = neg
        reg['gear/intf_oper/__rshift__'] = shr_or_code
        reg['gear/intf_oper/__sub__'] = sub
        reg['gear/intf_oper/__xor__'] = xor


@flow_visitor
class RemoveRecode(HDLGearHierVisitor):
    def code(self, node):
        pout = node.out_ports[0]
        pin = node.in_ports[0]

        if getattr(pin.dtype, 'width', 0) == getattr(pout.dtype, 'width', 0):
            node.bypass()


class HDLCodePlugin(VGenPlugin, SVGenPlugin):
    @classmethod
    def bind(cls):
        reg['vgen/flow'].insert(0, RemoveRecode)
        reg['svgen/flow'].insert(0, RemoveRecode)



from .find import find
from pygears import Intf
from pygears.lib import dreg as _dreg, decouple as _decouple


def pipeline(port_name):
    insert_module(port_name, _dreg)


def decouple(port_name, depth=2):
    insert_module(port_name, _decouple(depth=depth))


def insert_module(port_name, module):
    port = find(port_name)

    post_intf = port.consumer
    post_intf.disconnect(port)
    if hasattr(post_intf, 'var_name'):
        post_intf.var_name += '_' + module.definition.__name__

    pre_intf = Intf(post_intf.dtype)
    pre_intf.source(port)

    module(pre_intf, intfs=[post_intf])



from pygears.typing.visitor import TypingVisitorBase
from pygears.typing import bitw, is_type


class SVGenTypeVisitor(TypingVisitorBase):
    struct_str = "typedef struct packed {"
    union_str = "typedef union packed {"

    def __init__(self, name, basic_type='logic', depth=4):
        self.struct_array = []
        self.context = name
        self.basic_type = basic_type
        self.max_depth = depth
        self.depth = -1

    def visit(self, type_, field):
        if self.depth >= self.max_depth:
            return f'{self.basic_type} [{type_.width-1}:0]'

        self.depth += 1
        type_declaration = super().visit(type_, field)
        self.depth -= 1
        return type_declaration

    def visit_Int(self, type_, field):
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} signed [{type_.width-1}:0] {self.context}_t; // {type_}\n')
        return f'{self.basic_type} signed [{type_.width-1}:0]'

    def visit_Bool(self, type_, field):
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} {self.context}_t; // {type_}')
        return f'{self.basic_type}'

    def visit_Uint(self, type_, field):
        if (type_.width == 0):
            return None
        if (not self.depth):
            self.struct_array.append(
                f'typedef {self.basic_type} [{type_.width-1}:0] {self.context}_t; // {type_}\n')
        return f'{self.basic_type} [{type_.width-1}:0]'

    visit_Ufixp = visit_Uint
    visit_Fixp = visit_Int

    def visit_Unit(self, type_, field):
        return None

    def visit_Union(self, type_, field):
        struct_fields = []
        max_len = 0
        high_parent_context = self.context
        middle_parent_context = f'{high_parent_context}_data'
        low_parent_context = f'{middle_parent_context}_data'

        for t in (type_.types):
            if (t.width > max_len):
                max_len = t.width

        # if self.depth < self.max_depth:
        if False:
            for i, t in reversed(list(enumerate(type_.args))):
                field_tmp = f'f{i}'
                struct_fields.append(f'{self.struct_str} // ({t}, u?)')
                self.context = f'{low_parent_context}_{field_tmp}'
                type_declaration = self.visit(t, field_tmp)
                if (t.width < max_len):
                    struct_fields.append(
                        f'    {self.basic_type} [{t.width-1}:0] dummy; // u{t.width}')
                if (type_declaration is not None):
                    struct_fields.append(f'    {type_declaration} {field_tmp}; // {t}')
                struct_fields.append(f'}} {middle_parent_context}_{field_tmp}_t;\n')

            # create union
            struct_fields.append(f'{self.union_str} // {type_}')
            for i, t in reversed(list(enumerate(type_.args))):
                field_tmp = f'f{i}'
                struct_fields.append(
                    f'    {middle_parent_context}_{field_tmp}_t {field_tmp}; // ({t}, u?)')
            struct_fields.append(f'}} {middle_parent_context}_t;\n')

        # create struct
        struct_fields.append(f'{self.struct_str} // {type_}')
        struct_fields.append(
            f'    {self.basic_type} [{bitw(len(type_.args)-1)-1}:0] ctrl; // u{bitw(len(type_.args)-1)}'
        )

        if type_.data.width > 0:
            struct_fields.append(
                f'    {self.basic_type} [{type_.data.width-1}:0] data; // {type_.data}')

        struct_fields.append(f'}} {high_parent_context}_t;\n')

        self.struct_array.append('\n'.join(struct_fields))

        self.context = high_parent_context

        return f'{high_parent_context}_t'

    def visit_Queue(self, type_, field):
        struct_fields = []
        parent_context = self.context
        struct_fields.append(f'{self.struct_str} // {type_}')
        struct_fields.append(
            f'    {self.basic_type} [{type_.args[1]-1}:0] eot; // u{type_.args[1]}')
        self.context = f'{parent_context}_data'
        type_declaration = self.visit(type_.args[0], type_.fields[0])

        if (type_declaration is not None):
            struct_fields.append(f'    {type_declaration} data; // {type_.args[0]}')

        struct_fields.append(f'}} {parent_context}_t;\n')
        self.struct_array.append('\n'.join(struct_fields))

        self.context = parent_context

        return f'{parent_context}_t'

    def visit_Tuple(self, type_, field):
        if self.depth == self.max_depth:
            return self.visit_Uint(type_, field)

        struct_fields = []
        parent_context = self.context

        struct_fields.append(f'{self.struct_str} // {type_}')
        for t, f in zip(reversed(type_.args), reversed(type_.fields)):
            self.context = f'{parent_context}_{f}'
            type_declaration = self.visit(t, f)
            if type_declaration:
                struct_fields.append(f'    {type_declaration} {f}; // {t}')

        struct_fields.append(f'}} {parent_context}_t;\n')
        self.struct_array.append('\n'.join(struct_fields))

        self.context = parent_context

        return f'{parent_context}_t'

    def visit_Array(self, type_, field):
        if not self.depth:
            struct_fields = []
            parent_context = self.context
            self.context = f'{parent_context}_data'

        type_declaration = self.visit(type_.args[0], type_.fields[0])

        split_type = type_declaration.split(" ")
        if 'signed' in split_type:
            insert_idx = split_type.index('signed') + 1
        else:
            insert_idx = 1

        split_type.insert(insert_idx, f'[{int(type_.args[1])-1}:0]')
        type_declaration = ' '.join(split_type)

        if not self.depth:
            struct_fields.append(
                f'typedef {type_declaration} {parent_context}_t; // {type_}\n')
            self.struct_array.append('\n'.join(struct_fields))
            self.context = parent_context
            return f'{parent_context}_t'

        return f'{type_declaration}'


def svgen_typedef(dtype, name, depth=4):
    # TODO: if a variable is called "wr", of the type that has a composite
    # field "data", than struct for that field will be called "wr_data_t". If
    # there is a variable "wr_data" in same module, it will also have the type
    # called "wr_data_t". This can be in conflict when the type definitions are
    # added for verilator debugging. Maybe think of a different scheme to name
    # subtypes
    if isinstance(dtype, str):
        return f'typedef {dtype} {name}_t;'

    assert is_type(dtype)

    if dtype.width == 0:
        return f'typedef logic [0:0] {name}_t;'

    vis = SVGenTypeVisitor(name, depth=depth)
    vis.visit(type_=dtype, field=name)
    return '\n'.join(vis.struct_array)



class TypingVisitorBase:
    def visit(self, type_, field=None, **kwds):
        for c in type_.mro():
            visit_func_name = f'visit_{c.__name__}'

            if hasattr(self, visit_func_name):
                return getattr(self, visit_func_name)(type_, field, **kwds)

        else:
            return self.visit_default(type_, field, **kwds)

    def visit_Union(self, type_, field, **kwds):
        for t, f in zip(type_.types, type_.fields):
            self.visit(t, f)

    def visit_Int(self, type_, field, **kwds):
        pass

    def visit_Integer(self, type_, field, **kwds):
        pass

    def visit_Bool(self, type_, field, **kwds):
        pass

    def visit_Uint(self, type_, field, **kwds):
        pass

    def visit_Ufixp(self, type_, field, **kwds):
        pass

    def visit_Fixp(self, type_, field, **kwds):
        pass

    def visit_default(self, type_, field, **kwds):
        if hasattr(type_, 'fields'):
            return {
                f: self.visit(t, f, **kwds)
                for t, f in zip(type_, type_.fields)
            }
        else:
            try:
                return tuple(self.visit(t, **kwds) for t in type_)
            except TypeError:
                pass



from pygears import reg
from pygears.hls import ir, is_intf_id
from pygears.hls import Context, HDLVisitor, Scope
from pygears.typing import Bool, typeof
from pygears.core.port import HDLProducer, HDLConsumer
from pygears.core.gear import InSig, OutSig
from dataclasses import dataclass, field
from typing import List
import itertools
from pygears.hdl.sv.v.accessors import rewrite

from .util import svgen_typedef
from .v.util import vgen_signal, vgen_intf

# REG_TEMPLATE = """
# always @(posedge clk) begin
#     if(rst | _rst_cond) begin
#         {0} <= {1};
#     end else if ({0}_en && cycle_done) begin
#         {0} <= {0}_next;
#     end
# end
# """
REG_TEMPLATE = """
always @(posedge clk) begin
    if (rst || (({2}) && _state_en)) begin
        {0} <= {1};
    end else if (_state_en) begin
        {0} <= {0}_next;
    end
end
"""

REG_TEMPLATE_NO_RST_COND = """
always @(posedge clk) begin
    if (rst) begin
        {0} <= {1};
    end else if (_state_en) begin
        {0} <= {0}_next;
    end
end
"""


class HDLWriter:
    def __init__(self, indent=0):
        self.indent = indent
        self.lines = []

    def line(self, line=''):
        if not line:
            self.lines.append('')
        else:
            self.lines.append(f'{" "*self.indent}{line}')

    def block(self, block):
        for line in block.split('\n'):
            self.line(line)

    def __str__(self):
        return '\n'.join(self.lines)


res_true = ir.ResExpr(Bool(True))


@dataclass
class BlockLines:
    header: List = field(default_factory=list)
    content: List = field(default_factory=list)
    footer: List = field(default_factory=list)


def is_reg_id(expr):
    return (isinstance(expr, ir.Name) and isinstance(expr.obj, ir.Variable) and expr.obj.reg)


class SVCompiler(HDLVisitor):
    def __init__(self, ctx, var, writer, selected, lang, aux_funcs=None):
        self.ctx = ctx
        self.writer = writer
        self.var = var
        self.selected = selected
        self.block_lines = []
        self.block_stack = []
        self.defaults = Scope()
        self.lang = lang

        if self.lang == 'sv':
            from .sv_expression import svexpr
            self.separator = '.'
            self.svexpr = svexpr
        else:
            from .v.v_expression import vexpr
            self.separator = '.'
            self.svexpr = vexpr

        if aux_funcs is None:
            aux_funcs = {}

        self.aux_funcs = aux_funcs

    @property
    def cur_block_lines(self):
        return self.block_lines[-1]

    def trim_cur_block(self):
        content = []
        for c in self.cur_block_lines.content:
            if isinstance(c, BlockLines) and not c.content:
                continue

            content.append(c)

        self.cur_block_lines.content = content

    def prepend(self, line):
        self.cur_block_lines.content.insert(0, line)

    def write(self, line):
        self.cur_block_lines.content.append(line)

    def header(self, line):
        self.cur_block_lines.header.append(line)

    def footer(self, line):
        self.cur_block_lines.footer.append(line)

    def enter_block(self, block):
        self.defaults.subscope()
        self.block_stack.append(block)
        bl = BlockLines()
        self.block_lines.append(bl)

        if not isinstance(block, ir.HDLBlock):
            return

    def exit_block(self, block=None):
        self.defaults.upscope()

        bl = self.block_lines.pop()

        if isinstance(block, ir.Branch):
            maybe_else = 'else ' if getattr(block, 'else_branch', False) else ''
            # in_cond = ir.BinOpExpr((block.in_cond, block.opt_in_cond),
            #                        ir.opc.And)

            in_cond = block.test

            if in_cond != res_true:
                in_cond_val = self.svexpr(in_cond, self.aux_funcs)

                bl.header.append(f'{maybe_else}if ({in_cond_val}) begin')
            elif maybe_else:
                bl.header.append(f'else begin')

        self.write(bl)

        self.block_stack.pop()

    def handle_defaults(self, target, stmt):
        if target not in self.defaults:
            self.defaults.items[target] = stmt
        elif stmt != self.defaults[target]:
            self.defaults[target] = stmt
            self.write(stmt)

    def attr(self, *args):
        return self.separator.join(args)

    def _assign_value(self, target, val):
        if isinstance(target, ir.SubscriptExpr):
            base_target = target.val
        elif isinstance(target, ir.Name):
            base_target = target
        elif isinstance(target, ir.Component):
            base_target = target.val
        elif isinstance(target, ir.ConcatExpr):
            for i, t in enumerate(target.operands):
                self._assign_value(t, ir.SubscriptExpr(val, ir.ResExpr(i)))

            return
        else:
            raise Exception

        if is_reg_id(base_target):

            if not self.selected(base_target):
                return

            name = self.svexpr(target, self.aux_funcs)

            val = self.svexpr(val, self.aux_funcs)
            if val is None:
                return

            svstmt = f"{name} = {val}"

            # if this is only partial assignement, the order of assignment then
            # matters and this cannot be a default
            if isinstance(target, ir.SubscriptExpr):
                self.write(svstmt)
            else:
                self.handle_defaults(name, svstmt)

            if name == '_state_next':
                self.handle_defaults("_state_en", "_state_en = 1")

            return

        if target.dtype is OutSig:
            if self.selected(target):
                name = target.obj.val.name
                svval = self.svexpr(val, self.aux_funcs)
                svstmt = f"{name} = {svval}"
                self.handle_defaults(name, svstmt)
            return
        elif isinstance(target, ir.Component):
            name = self.svexpr(base_target, self.aux_funcs)
            if target.field == 'data':
                if self.selected(base_target):
                    self.handle_defaults(self.attr(name, 'valid'),
                                         f"{self.attr(name, 'valid')} = 1")
                    val = self.svexpr(val, self.aux_funcs)
                    if val is not None:
                        svstmt = f"{name}_s = {val}"
                        self.handle_defaults(name, svstmt)
            elif target.field == 'ready' and self.selected(base_target):
                self.handle_defaults(self.attr(name, 'ready'), f"{self.attr(name, 'ready')} = 1")

            return
        elif is_intf_id(target):
            if is_intf_id(val) and self.selected(val):
                name = self.svexpr(base_target, self.aux_funcs)
                val_name = self.svexpr(val, self.aux_funcs)
                svstmt = f"{name}_s = {val_name}_s"
                self.handle_defaults(name, svstmt)
                self.write(f"{self.attr(name, 'valid')} = {self.attr(val_name, 'valid')}")
                self.write(f"{self.attr(val_name, 'ready')} = {self.attr(name, 'ready')}")

            return

        if not self.selected(base_target):
            return

        if val.dtype is None:
            return

        target = self.svexpr(target, self.aux_funcs)

        if target is None:
            return

        val = self.svexpr(val, self.aux_funcs)
        if val is None:
            return

        svstmt = f"{target} = {val}"

        self.handle_defaults(target, svstmt)

    def AssertValue(self, node):
        self.write(f'assert ({self.svexpr(node.val.test, self.aux_funcs)})')
        self.write(f'else $error("{node.val.msg}");')

    def AssignValue(self, node):
        self._assign_value(node.target, node.val)

    def list_initials(self):
        for name, obj in self.ctx.scope.items():
            if not isinstance(obj, ir.Variable):
                continue

            if typeof(obj.dtype, ir.IntfType):
                # if isinstance(obj.val.producer, HDLProducer):
                if (self.selected(self.ctx.ref(name, ctx='store'))
                        and obj.dtype.direction == ir.IntfType.iout):
                    yield self.attr(name, 'valid'), '0'
                # elif len(obj.val.consumers) == 1 and isinstance(obj.val.consumers[0], HDLConsumer):
                elif self.selected(self.ctx.ref(name, ctx='ready')):
                    if not is_port_intf(name, self.ctx):
                        yield self.attr(name, 'ready'), f"{self.attr(name, 'valid')} ? 0 : 1'bx"

            elif obj.reg:
                target = self.ctx.ref(name)
                if self.selected(target):
                    if name == '_state':
                        yield '_state_en', '0'

                    yield (f'{self.svexpr(self.ctx.ref(name, ctx="store"), self.aux_funcs)}',
                           f'{self.svexpr(self.ctx.ref(name), self.aux_funcs)}')
            else:
                pass

    def Module(self, node):
        self.block_lines.append(BlockLines())

        self.header(f'// Comb block for: {self.var}')
        self.header(f'always_comb begin')

        for target, expr in self.list_initials():
            self.handle_defaults(target, f"{target} = {expr}")
            # self.prepend(f"{target} = {stmt}")
            # self.defaults[target] = stmt

        self.BaseBlock(node)

        self.trim_cur_block()

        for target, svstmt in reversed(list(self.defaults.items.items())):
            self.prepend(svstmt)

        self.footer('end')

    def FuncReturn(self, node):
        retval = self.svexpr(node.expr, self.aux_funcs)

        if self.lang == 'sv':
            if retval is None:
                self.write('return')
            else:
                self.write(f"return {self.svexpr(node.expr, self.aux_funcs)}")
        else:
            if retval is not None:
                self.write(
                    f"{self.svexpr(node.func.name, self.aux_funcs)} = {self.svexpr(node.expr, self.aux_funcs)}"
                )

    def FuncBlock(self, node):
        self.block_lines.append(BlockLines())

        self.header('')

        self.BaseBlock(node)

        self.footer(f'endfunction')
        self.footer('')

    def LoopBlock(self, node):
        raise Exception()

    def BaseBlock(self, node):
        self.enter_block(node)

        for stmt in node.stmts:
            self.visit(stmt)

        self.trim_cur_block()

        self.exit_block(node)

    def HDLBlock(self, node):
        self.enter_block(node)

        for i, b in enumerate(node.branches):
            if i > 0:
                b.else_branch = True
            else:
                b.else_branch = False

            self.visit(b)

        content = []
        for c in reversed(self.cur_block_lines.content):
            if not content and not c.content:
                continue

            content.insert(0, c)

        self.cur_block_lines.content = content

        self.exit_block(node)


gear_module_template = """
{%- import 'snippet.j2' as snippet -%}

{% call snippet.gear_module(module_name, intfs, comment, sigs) %}

{{svlines|indent(4,True)}}

{%- endcall %}
"""


def write_block(block, writer):
    for h in block.header:
        writer.line(h)

    if block.header:
        writer.indent += 4

    for c in block.content:
        if isinstance(c, str):
            writer.line(c + ';')
        else:
            write_block(c, writer)

    if block.header:
        writer.indent -= 4
        if block.footer and block.footer[0] == 'endfunction':
            writer.line('endfunction')
        else:
            writer.line('end')


def typedef_or_inline(writer, dtype, name):
    res = svgen_typedef(dtype, name)
    if "\n" not in res[:-1]:
        return res.partition(' ')[-1].rpartition(';')[0].rpartition(' ')[0]

    writer.block(res)
    return f'{name}_t'


def svcompile(hdl_stmts, ctx, title, selected, lang, aux_funcs=None):
    writer = HDLWriter()
    v = SVCompiler(ctx, title, writer, selected=selected, lang=lang, aux_funcs=aux_funcs)
    v.visit(hdl_stmts)

    if not v.block_lines[0].content:
        return ''

    write_block(v.block_lines[0], writer)
    writer.line()
    return str(writer)


# TODO: Why do we need this check, can we generalize this for any variable?
def is_top_port_intf(name, ctx):
    for p in ctx.gear.in_ports + ctx.gear.out_ports:
        if p.basename == name:
            return p
    else:
        return None


def is_port_intf(name, ctx):
    for m in ctx.submodules:
        for p in m.in_ports:
            if p.name == name:
                return p

    for p in ctx.gear.in_ports:
        if p.basename == name:
            if p.consumer:
                return p
            else:
                return None

    for p in ctx.gear.out_ports:
        if p.basename == name:
            return p


def write_declarations(ctx, subsvmods, template_env):
    writer = HDLWriter()

    lang = template_env.lang

    if lang == 'sv':
        from .sv_expression import svexpr
        sep = '.'
        exprgen = svexpr
    else:
        from .v.v_expression import vexpr
        sep = '.'
        exprgen = vexpr

    # TODO: Fix not to require this check
    if '_state_en' not in ctx.scope:
        writer.line(f'logic _state_en;')

    for name, expr in ctx.regs.items():
        name = exprgen(ctx.ref(name))

        if lang == 'sv':
            name_t = typedef_or_inline(writer, expr.dtype, name)
            writer.line(f'{name_t} {name}, {name}_next;')
        else:
            writer.block(vgen_signal(expr.dtype, 'reg', f'{name}_next', 'output', hier=False))
            writer.block(vgen_signal(expr.dtype, 'reg', name, 'output', hier=False))

        writer.line()

    for name, expr in ctx.intfs.items():
        dtype = expr.dtype.dtype
        if is_top_port_intf(name, ctx) is None:
            if lang == 'sv':
                writer.line(f'dti#({dtype.width}) {name}();')
            else:
                writer.line(f'reg {name}_ready;')
                writer.line(f'reg {name}_valid;')
                writer.line(vgen_signal(dtype, 'reg', f'{name}_data', 'output', False))

        if lang == 'sv':
            name_t = typedef_or_inline(writer, dtype, name)
            writer.line(f'{name_t} {name}_s;')

            if expr.dtype.direction == ir.IntfType.iin:
                writer.line(f'assign {name}_s = {sep.join([name, "data"])};')
            else:
                writer.line(f'assign {sep.join([name, "data"])} = {name}_s;')

        else:
            if expr.dtype.direction == ir.IntfType.iin:
                writer.block(vgen_signal(dtype, 'reg', f'{name}_s', 'input', False))
                writer.line(f"assign {name}_s = {name}_data;")
            else:
                writer.block(vgen_signal(dtype, 'reg', f'{name}_s', 'output', False))
                writer.line(f"assign {name}_data = {name}_s;")

        writer.line()

    for name, expr in ctx.variables.items():
        if expr.dtype is None or typeof(expr.dtype, ir.IntfType) or expr.reg:
            continue

        if lang == 'sv':
            name_t = typedef_or_inline(writer, expr.dtype, name)
            writer.line(f'{name_t} {name};')
        else:
            writer.block(vgen_signal(expr.dtype, 'reg', name, 'input', hier=False))

        writer.line()

    for c, s in zip(ctx.submodules, subsvmods):
        port_map = {}
        for intf, p in itertools.chain(zip(c.in_ports, c.gear.in_ports),
                                       zip(c.out_ports, c.gear.out_ports)):
            if lang == 'sv':
                port_map[p.basename] = intf.name
            else:
                port_map[p.basename] = (intf.name, None, None)

        writer.block(s.get_inst(template_env, port_map))

    if ctx.regs:
        writer.line(f'initial begin')
        for name, expr in ctx.regs.items():
            if not isinstance(expr.val, ir.ResExpr):
                continue

            writer.line(f"    {exprgen(ctx.ref(name))} = {exprgen(expr.val)};")

        writer.line(f'end')

    return str(writer)


def write_module(ctx: Context, hdl, writer, subsvmods, funcs, template_env, config=None):
    if config is None:
        config = {}

    lang = template_env.lang

    aux_funcs = {}

    if lang == 'sv':
        from .sv_expression import svexpr
        sep = '.'
        exprgen = svexpr
    else:
        from .v.v_expression import vexpr
        sep = '.'
        exprgen = vexpr

    for f_hdl, f_ctx in funcs:
        size = ''
        if f_hdl.ret_dtype.width > 0:
            size = f'[{f_hdl.ret_dtype.width-1}:0]'

        if getattr(f_hdl.ret_dtype, 'signed', False):
            size = f'signed {size}'

        writer.line(f'function {size} {f_hdl.name};')

        writer.indent += 4

        for name, expr in f_ctx.variables.items():
            if expr.dtype is None:
                continue

            if name in f_ctx.signature:
                continue

            if lang == 'sv':
                name_t = typedef_or_inline(writer, expr.dtype, name)
                writer.line(f'{name_t} {name};')
            else:
                writer.block(vgen_signal(expr.dtype, 'reg', name, 'input', hier=False))

            writer.line()

        if lang == 'sv':
            for name, dtype in f_ctx.signature.items():
                name_t = typedef_or_inline(writer, dtype, name)
                writer.line(f'input {name_t} {name};')
                writer.line()
        else:
            for name, dtype in f_ctx.signature.items():
                writer.block(vgen_signal(dtype, 'input', name, 'input', hier=False))
                writer.line()

            for name, dtype in f_ctx.signature.items():
                tmp = vgen_signal(dtype, 'reg', name, 'input', hier=False)
                writer.block('\n'.join(l for l in tmp.split('\n')[1:] if l.startswith('reg')))

            for name, dtype in f_ctx.signature.items():
                tmp = vgen_signal(dtype, 'reg', name, 'input', hier=False)
                for l in tmp.split('\n')[1:]:
                    if l.startswith('reg'):
                        continue

                    if l.startswith('assign'):
                        l = l[7:]

                    writer.line(l)

        writer.indent -= 4

        blk = svcompile(f_hdl, f_ctx, '', selected=lambda x: True, lang=lang, aux_funcs=aux_funcs)

        if lang == 'v':
            blk = vrewrite(f_ctx, blk)

        writer.line(blk)

    blk = write_declarations(ctx, subsvmods, template_env)

    for name, expr in ctx.regs.items():
        init = "'x" if expr.val is None else exprgen(expr.val)
        if not ctx.reset_states.get(name, None):
            blk += REG_TEMPLATE_NO_RST_COND.format(exprgen(ctx.ref(name)), init)
        else:
            rst_states = ctx.reset_states[name]
            # TODO: Find a better way to infer when the state register has been removed
            if '_rst_cond' not in ctx.scope:
                rst_expr = ' || '.join([f'(_state_next == {exprgen(s)})' for s in rst_states])
            else:
                rst_expr = ' || '.join([f'_rst_cond[{s}]' for s in rst_states])

            blk += REG_TEMPLATE.format(exprgen(ctx.ref(name)), init, rst_expr)

    for name, expr in ctx.regs.items():
        blk += svcompile(hdl,
                         ctx,
                         name,
                         selected=lambda x: x.obj == expr,
                         lang=lang,
                         aux_funcs=aux_funcs)

    for name, expr in ctx.variables.items():
        var_blk = svcompile(hdl,
                         ctx,
                         name,
                         selected=lambda x: x.obj == expr,
                         lang=lang,
                         aux_funcs=aux_funcs)

        if typeof(expr.dtype, ir.IntfType) and any(expr.name == p.basename for p in ctx.gear.in_ports):
            # TODO: Make this transformation somewhere else
            if len(ctx.gear.in_ports) == len(ctx.gear.out_ports) == 1:
                pin = ctx.gear.in_ports[0].basename
                pout = ctx.gear.out_ports[0].basename

                template = (
                    f'if ({pin}.valid) begin\n'
                    f'    if ({pout}.ready) begin\n'
                    f'        {pin}.ready = 1;\n'
                    f'    end\n'
                    f'end'
                )

                if all(t.strip() == v.strip() for t, v in zip(template.split('\n'), var_blk.split('\n')[3:8])):
                    var_blk = f'    assign {pin}.ready = {pout}.ready;'

        blk += var_blk

    for name, expr in ctx.signals.items():
        blk += svcompile(hdl,
                         ctx,
                         name,
                         selected=lambda x: x.obj == expr,
                         lang=lang,
                         aux_funcs=aux_funcs)

    # for name, expr in ctx.intfs.items():
    #     blk += svcompile(hdl,
    #                      ctx,
    #                      name,
    #                      selected=lambda x: x.name == name,
    #                      lang=lang,
    #                      aux_funcs=aux_funcs)

    if lang == 'v':
        blk = vrewrite(ctx, blk)

    writer.line(blk)

    writer.lines[0:0] = aux_funcs.values()


def vrewrite(ctx, body):
    index = {}
    for name, v in ctx.scope.items():
        if not isinstance(v, ir.Variable):
            continue

        index[name] = v.dtype
        if typeof(v.dtype, ir.IntfType):
            index[f'{name}_s'] = v.dtype.dtype
        else:
            if v.reg:
                index[f'{name}_next'] = v.dtype

    return rewrite(body, index)


def compile_gear_body(gear, outdir, template_env):
    # ctx, hdl_ast = parse_gear_body(gear)
    from pygears.hls.translate import translate_gear
    ctx, hdl_ast = translate_gear(gear)

    subsvmods = []
    if ctx.submodules:
        from pygears.hdl import hdlgen
        hdlgen_map = reg['hdlgen/map']
        for c in ctx.submodules:
            rtl_top = hdlgen(c.gear, outdir=outdir, generate=False)
            svmod = hdlgen_map[rtl_top]
            subsvmods.append(svmod)

    funcs = []

    def _get_funcs_rec(block):
        for f_ast, f_ctx in block.funcs:
            funcs.append((f_ast, f_ctx))
            _get_funcs_rec(f_ast)

    _get_funcs_rec(hdl_ast)

    writer = HDLWriter()
    write_module(ctx,
                 hdl_ast,
                 writer,
                 subsvmods,
                 funcs,
                 template_env,
                 config=gear.meta_kwds.get('hdl', {}))

    return '\n'.join(writer.lines), subsvmods


def compile_gear(gear, template_env, module_name, outdir, comment=None, attrib=None):
    # TODO: Harden the case where local variable shadows a global one
    context = {
        'module_name': module_name,
        'intfs': template_env.port_intfs(gear),
        'sigs': gear.meta_kwds['signals'],
        'params': gear.params,
        'comment': {
            'comment': comment,
            'attrib': [] if attrib is None else attrib
        }
    }

    context['svlines'], subsvmods = compile_gear_body(gear, outdir, template_env)

    return template_env.render_string(gear_module_template, context), subsvmods



from pygears import gear
from pygears.typing import Tuple, Uint, Any


@gear
async def dispatch(din: Tuple['data':Any, 'ctrl':Uint]) -> b'(din["data"], ) * din["ctrl"].width':
    async with din as (data, ctrl):
        if ctrl != 0:
            dout = [data if c else None for c in ctrl]
            yield tuple(dout)



from .intf import Intf
from pygears.conf import core_log
from pygears.core.graph import get_source_producer, get_producer_port
from pygears.core.hier_node import find_unique_names
from .port import InPort, OutPort
from .gear import OutSig, InSig


def connect_to_existing_parent_in_port(in_port):
    parent = in_port.gear.parent
    in_intf = in_port.producer
    src_intf = get_source_producer(in_port)

    for parent_port in parent.in_ports:
        if src_intf is get_source_producer(parent_port):
            in_intf.disconnect(in_port)
            parent_port.consumer.connect(in_port)
            return True

    return False


def connect_to_existing_parent_out_port(out_port, cons_port):
    parent = out_port.gear.parent
    out_intf = out_port.consumer

    for parent_port in parent.out_ports:
        if out_intf is parent_port.consumer:
            out_intf.disconnect(cons_port)
            parent_port.consumer.connect(cons_port)
            return True

    return False


# TODO: Dangling reporting not working anymore
def report_out_dangling(port):
    src_intf = get_source_producer(port)
    p = src_intf.consumers[0]

    if hasattr(src_intf, 'var_name'):
        core_log().warning(f'Interface "{p.gear.name}/{src_intf.var_name}" left dangling.')
    else:
        path = []
        while True:
            g = p.gear

            if hasattr(p.consumer, 'var_name'):
                path.append(f'{g.parent.name}/{p.consumer.var_name}')
            else:
                path.append(p.name)

            if len(g.in_ports) != 1 or len(g.out_ports) != 1:
                break

            p = get_producer_port(g.in_ports[0])

        path = ' -> '.join(reversed(path))

        core_log().warning(f'Interface "{path}" left dangling.')


def channel_out_port(gear_inst, out_port):
    out_parent_cons = []

    out_intf = out_port.consumer
    for cons_port in out_intf.consumers:
        cons_gear = cons_port.gear

        if gear_inst.parent.has_descendent(cons_gear):
            continue

        if connect_to_existing_parent_out_port(out_port, cons_port):
            continue

        out_parent_cons.append(cons_port)

    if out_intf.consumers and not out_parent_cons:
        return

    basename = getattr(out_intf, 'var_name', out_port.basename)
    new_name = next(
        find_unique_names(
            [basename] +
            [p.basename for p in (gear_inst.parent.out_ports + gear_inst.parent.in_ports)]))

    if new_name:
        basename = new_name

    parent_port = OutPort(gear_inst.parent, len(gear_inst.parent.out_ports), basename)

    gear_inst.parent.out_ports.append(parent_port)

    in_intf = Intf(out_intf.dtype)
    out_intf.source(parent_port)
    in_intf.source(out_port)
    in_intf.connect(parent_port)

    for p in out_intf.consumers:
        if p in out_parent_cons:
            continue

        out_intf.disconnect(p)
        in_intf.connect(p)


def channel_in_port(gear_inst, in_port):
    in_intf = in_port.producer
    prod_port = in_intf.producer

    if prod_port:
        if gear_inst.parent.has_descendent(prod_port.gear):
            return

    if connect_to_existing_parent_in_port(in_port):
        return

    basename = getattr(in_intf, 'var_name', in_port.basename)
    new_name = next(
        find_unique_names(
            [basename] +
            [p.basename for p in (gear_inst.parent.out_ports + gear_inst.parent.in_ports)]))

    if new_name:
        basename = new_name

    parent_port = InPort(gear_inst.parent, len(gear_inst.parent.in_ports), basename)

    gear_inst.parent.in_ports.append(parent_port)

    in_intf.disconnect(in_port)
    in_intf.connect(parent_port)

    gear_in_intf = Intf(in_intf.dtype)
    gear_in_intf.source(parent_port)
    gear_in_intf.connect(in_port)


def is_driven_by_node(node, name):
    for s in node.meta_kwds['signals']:
        if isinstance(s, OutSig):
            if s.name in node.params['sigmap']:
                if name == node.params['sigmap'][s.name]:
                    return s
            else:
                if name == s.name:
                    return s
    else:
        return None


def find_signal_driver_port(parent, name):
    for node in parent.child:
        if is_driven_by_node(node, name):
            return True

    for s in parent.meta_kwds['signals']:
        if isinstance(s, InSig):
            if s.name == name:
                return True

    return False


def channel_interfaces(gear_inst):
    for in_port in gear_inst.in_ports:
        channel_in_port(gear_inst, in_port)

    for out_port in gear_inst.out_ports:
        channel_out_port(gear_inst, out_port)

    for s in gear_inst.meta_kwds['signals']:
        if s.name not in gear_inst.params['sigmap']:
            gear_inst.params['sigmap'] = gear_inst.params['sigmap'].copy()
            gear_inst.params['sigmap'][s.name] = s.name

    for s in gear_inst.meta_kwds['signals']:
        if isinstance(s, InSig):
            sig_name = gear_inst.params['sigmap'].get(s.name, s.name)

            if not sig_name.isidentifier():
                continue

            # sig_name = s.name

            if not find_signal_driver_port(gear_inst.parent, sig_name):
                if not isinstance(gear_inst.parent.meta_kwds['signals'], list):
                    gear_inst.parent.meta_kwds['signals'] = list(
                        gear_inst.parent.meta_kwds['signals'])

                gear_inst.parent.meta_kwds['signals'] = gear_inst.parent.meta_kwds['signals'].copy()
                gear_inst.parent.meta_kwds['signals'].append(s)



import ctypes
import shutil
import atexit
import os
import subprocess
from string import Template
from pygears import Intf

import jinja2

from pygears import reg, find
from pygears.sim import log
from pygears.sim.c_drv import CInputDrv, COutputDrv
from pygears.sim.modules.cosim_base import CosimBase
from pygears.hdl import hdlgen, list_hdl_files
from pygears.util.fileio import save_file
from pygears.core.port import InPort, OutPort
from .cosim_port import InCosimPort, OutCosimPort

signal_spy_connect_t = Template("""
/*verilator tracing_on*/
${intf_name}_t ${intf_name}_data;
logic ${intf_name}_valid;
logic ${intf_name}_ready;
/*verilator tracing_off*/

assign ${intf_name}_data = ${conn_name}.data;
assign ${intf_name}_valid = ${conn_name}.valid;
assign ${intf_name}_ready = ${conn_name}.ready;
""")

signal_spy_connect_hide_interm_t = Template("""
/*verilator tracing_on*/
${intf_name}_t ${intf_name}_data;
logic ${intf_name}_valid;
logic ${intf_name}_ready;
/*verilator tracing_off*/

always_comb
    if (${conn_name}.valid)
        ${intf_name}_data = ${conn_name}.data;

assign ${intf_name}_valid = ${conn_name}.valid;
assign ${intf_name}_ready = ${conn_name}.ready;
""")

signal_spy_connect_no_struct = Template("""
/*verilator tracing_on*/
dti #(.W_DATA(${width})) _${intf_name}();
/*verilator tracing_off*/

assign _${intf_name}.data = ${conn_name}.data;
assign _${intf_name}.valid = ${conn_name}.valid;
assign _${intf_name}.ready = ${conn_name}.ready;
""")


class VerilatorCompileError(Exception):
    pass


# cosim('/top', 'verilator', build=True, rebuild=False)


def get_file_struct(top, outdir):
    from pygears.sim import cosim_build_dir

    name = cosim_build_dir(top)

    if outdir is None:
        outdir = reg['results-dir']

    outdir = os.path.abspath(os.path.join(outdir, name))
    objdir = os.path.join(outdir, 'obj_dir')

    dll_path = os.path.join(objdir, f'pygearslib')
    if os.name == 'nt':
        dll_path += '.exe'

    return {'name': name, 'outdir': outdir, 'objdir': objdir, 'dll_path': dll_path}


def make(objdir, top_name):
    ret = os.system(f'cd {objdir}; make -j -f V{top_name}.mk > make.log 2>&1')
    # TODO: Not much of a speedup f'cd {self.objdir}; make -j OPT_FAST="-Os -fno-stack-protector" -f V{self.top_name}.mk > make.log 2>&1')

    if ret != 0:
        raise VerilatorCompileError(f'Verilator compile error: {ret}. '
                                    f'Please inspect "{objdir}/make.log"')


def create_project_script(outdir, top, lang):
    hdl_files = list_hdl_files(top, outdir)
    with open(os.path.join(outdir, 'verilator.prj'), 'w') as f:
        for fn in hdl_files:
            f.write(f'{fn}\n')

    # with open(os.path.join(outdir, 'verilator.prj'), 'w') as f:
    #     for fn in reg['hdlgen/hdlmods'].values():
    #         f.write(f'{fn}\n')


def verilate(outdir, lang, top, top_name, tracing_enabled, rst=True, expand_data=True):
    # include = ' '.join([f'-I{os.path.abspath(p)}' for p in reg[f'{lang}gen/include']])

    # include += f' -I{outdir}'

    # files = f'{wrap_name}.{lang}'
    # create_project_script(outdir, top, lang)

    verilate_cmd = [
        f'cd {outdir};',
        'verilator -cc -CFLAGS -fpic -LDFLAGS -shared --exe',
        '-Wno-fatal',
        # TODO: Not much of a speedup: '-O3 --x-assign fast --x-initial fast --noassert',
        '-clk clk',
        f'--top-module {top_name}',
        '--trace --no-trace-params --trace-underscore' if tracing_enabled else '',
        '--trace-structs' if tracing_enabled and expand_data else '',
        '--x-initial unique' if not rst else '',
        '-o pygearslib',
        top_name,
        'sim_main.cpp',
    ]

    # '--x-assign unique' if not rst else '',

    with open(os.path.join(outdir, 'verilate.log'), 'w') as f:
        f.write(" ".join(verilate_cmd))
        f.write("\n")

    ret = subprocess.call(f'{" ".join(verilate_cmd)} >> verilate.log 2>&1', shell=True)

    if ret != 0:
        raise VerilatorCompileError(f'Verilator compile error: {ret}. '
                                    f'Please inspect "{outdir}/verilate.log"')


def build(top, outdir=None, postsynth=False, lang=None, rebuild=True, rst=True):
    if isinstance(top, str):
        top_name = top
        top = find(top)

        if top is None:
            raise Exception(f'No gear found on path: "{top_name}"')

    if lang is None:
        lang = reg['hdl/lang']

    if lang != 'v':
        postsynth = False
    else:
        pass
        # postsynth = True

    file_struct = get_file_struct(top, outdir)

    outdir = file_struct['outdir']

    if not rebuild and os.path.exists(file_struct['dll_path']):
        return

    log.info(f'Verilating "{top.name}" inside {outdir}...')
    shutil.rmtree(outdir, ignore_errors=True)

    if not reg['debug/expand_trace_data']:
        reg['svgen/spy_connection_template'] = signal_spy_connect_no_struct
    elif reg['debug/hide_interm_vals']:
        reg['svgen/spy_connection_template'] = signal_spy_connect_hide_interm_t
    else:
        reg['svgen/spy_connection_template'] = signal_spy_connect_t

    synth_src_dir = os.path.join(outdir, 'src')
    hdlgen(top,
           outdir=synth_src_dir if postsynth else outdir,
           generate=True,
           lang=lang,
           copy_files=True,
           toplang='v')

    hdlmod = reg['hdlgen/map'][top]

    if postsynth:
        # TODO: change this to the call of the toplevel synth function
        from pygears.hdl.yosys import synth
        synth(outdir=outdir,
              top=top,
              synthcmd='synth',
              srcdir=synth_src_dir,
              synthout=os.path.join(outdir, hdlmod.file_basename))

    tracing_enabled = bool(reg['debug/trace'])
    context = {
        'in_ports': top.in_ports,
        'out_ports': top.out_ports,
        'top_name': hdlmod.wrap_module_name,
        'tracing': tracing_enabled,
        'end_cycle_dump': reg['debug/trace_end_cycle_dump'],
        'rst': rst,
        'aux_clock': reg['sim/aux_clock']
    }

    jenv = jinja2.Environment(trim_blocks=True, lstrip_blocks=True)
    jenv.globals.update(int=int)
    jenv.loader = jinja2.FileSystemLoader([os.path.dirname(__file__)])
    c = jenv.get_template('sim_veriwrap.j2').render(context)
    save_file('sim_main.cpp', outdir, c)

    verilate(outdir,
             lang,
             top,
             hdlmod.wrap_module_name,
             tracing_enabled,
             rst=rst,
             expand_data=reg['debug/expand_trace_data'])

    make(file_struct['objdir'], hdlmod.wrap_module_name)


class SimVerilated(CosimBase):
    def __init__(self,
                 gear,
                 timeout=100,
                 rebuild=True,
                 vcd_fifo=False,
                 shmidcat=False,
                 postsynth=False,
                 outdir=None,
                 lang=None,
                 rst=True):

        super().__init__(gear, timeout=timeout)
        self.name = gear.name[1:].replace('/', '_')
        self.outdir = outdir
        self.rst = rst
        self.rebuild = rebuild
        self.top = gear
        self.lang = lang
        if self.lang is None:
            self.lang = reg['hdl/lang']

        self.trace_fn = None
        self.vcd_fifo = vcd_fifo
        self.shmidcat = shmidcat
        self.shmid_proc = None
        self.verilib = None
        self.finished = False

    def cycle(self):
        self.verilib.cycle()

    def forward(self):
        self.verilib.forward()

    def back(self):
        self.verilib.back()

    def setup(self):
        # TODO: When reusing existing verilated build, add test to check
        # whether verilated module is the same as the current one (Maybe hash check?)
        if self.rebuild:
            build(self.top, self.outdir, postsynth=False, lang=self.lang, rst=self.rst)
            log.info(f'Done')

        file_struct = get_file_struct(self.top, self.outdir)

        tracing_enabled = bool(reg['debug/trace'])
        if tracing_enabled:
            log.info(f"Debug: {reg['debug/trace']}")
            self.trace_fn = os.path.join(reg["results-dir"], f'{self.name}.vcd')
            try:
                subprocess.call(f"rm -f {self.trace_fn}", shell=True)
            except OSError:
                pass

            if self.vcd_fifo:
                subprocess.call(f"mkfifo {self.trace_fn}", shell=True)
            else:
                log.info(f'Verilator VCD dump to "{self.trace_fn}"')
        else:
            self.trace_fn = ''

        try:
            self.verilib = ctypes.CDLL(file_struct['dll_path'])
        except OSError:
            raise VerilatorCompileError(
                f'Verilator compiled library for the gear "{self.gear.name}"'
                f' not found at: "{file_struct["dll_path"]}"')

        self.finished = False
        atexit.register(self._finish)

        if self.shmidcat and tracing_enabled:
            self.shmid_proc = subprocess.Popen(f'shmidcat {self.trace_fn}',
                                               shell=True,
                                               stdout=subprocess.PIPE)

            # Wait for shmidcat to actually open the pipe, which is necessary
            # to happen prior to init of the verilator. If shmidcat does not
            # open the pipe, verilator will get stuck
            import time
            time.sleep(0.1)

        self.verilib.init.argtypes = [ctypes.c_char_p, ctypes.c_long]

        assert self.verilib.init(self.trace_fn.encode('utf8'), reg['sim/rand_seed']) != 0

        if self.shmid_proc:
            self.shmid = self.shmid_proc.stdout.readline().decode().strip()
            log.info(f'Verilator VCD dump to shared memory at 0x{self.shmid}')

        self.handlers = {}
        for cp in self.in_cosim_ports:
            self.handlers[cp.name] = CInputDrv(self.verilib, cp.port, cp.name)

        for cp in self.out_cosim_ports:
            self.handlers[cp.name] = COutputDrv(self.verilib, cp.port, cp.name)

        super().setup()

    def _finish(self):
        if not self.finished:

            if self.eval_needed:
                self.verilib.forward()
            self.verilib.cycle()

            self.finished = True
            self.handlers.clear()
            super()._finish()
            if self.shmid_proc:
                self.shmid_proc.terminate()

            if self.verilib:
                self.verilib.final()

            self.verilib = None

        atexit.unregister(self._finish)



import ast
from . import Context, ir, node_visitor, visit_ast
from .cast import resolve_cast_func
from .stmt import assign_targets, extend_stmts
from ..ir_utils import is_intf_id


def is_target_id(node):
    return isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store)


def merge_cond_alias_map(ir_node, ctx):
    alias_map = ctx.alias_map
    for n, v in ctx.closure_alias_maps[ir_node].items():
        if n in alias_map:
            if v is not alias_map[n] and v != alias_map[n]:
                alias_map[n] = ctx.ref(n)


# TODO: Revisit cast_return, maybe it can be more general
from pygears.typing import typeof, Queue, Tuple, Array


def cast_rec(node, t):
    if node == ir.ResExpr(None):
        return ir.ResExpr(ir.EmptyType[t]())
    elif node.dtype == t:
        return node
    elif typeof(node.dtype, ir.IntfType):
        # TODO: Do proper casting of interfaces
        return node
    elif not typeof(t, (Queue, Tuple, Array)):
        return resolve_cast_func(node, t)

    if isinstance(node, ir.ConcatExpr):
        ops = node.operands
    elif isinstance(node, ir.ResExpr):
        ops = [ir.ResExpr(v) for v in node.val]
    else:
        # TODO: Should this be done with indexing to ensure proper casting of fields?
        return node

    opout = []
    for i in range(len(ops)):
        opout.append(cast_rec(ops[i], t[i]))

    return ir.ConcatExpr(opout)


def cast_return(arg_nodes, out_ports):
    out_num = len(out_ports)

    # TODO: Reconsider this: Can we have a type inference mechanism?
    if out_num == 0:
        raise TypeError(f"Function must have a return type defined!")

    if isinstance(arg_nodes, (list, tuple)):
        assert len(arg_nodes) == out_num
        input_vars = arg_nodes
    elif isinstance(arg_nodes, ir.TupleExpr):
        input_vars = arg_nodes.val
    elif isinstance(arg_nodes, ir.Name) and out_num > 1:
        var = arg_nodes.obj
        assert len(var.dtype) == out_num
        input_vars = []
        for i in range(len(var.dtype)):
            input_vars.append(ir.SubscriptExpr(val=arg_nodes, index=ir.ResExpr(i)))
    else:
        if out_num == 1:
            input_vars = [arg_nodes]
        elif isinstance(arg_nodes, ir.TupleExpr):
            input_vars = arg_nodes.val
        elif isinstance(arg_nodes, ir.ConcatExpr):
            input_vars = arg_nodes.operands
        elif isinstance(arg_nodes, ir.ResExpr) and isinstance(arg_nodes.val, tuple):
            input_vars = [ir.ResExpr(v) for v in arg_nodes.val]
        else:
            breakpoint()
            raise Exception('Unsupported')

    args = []
    for arg, intf in zip(input_vars, out_ports):
        port_t = intf.dtype
        args.append(cast_rec(arg, port_t))

    return ir.TupleExpr(args)


@node_visitor(ast.Yield)
def parse_yield(node, ctx):
    yield_expr = visit_ast(node.value, ctx)

    try:
        ret = cast_return(yield_expr, ctx.gear.out_ports)
    except TypeError as e:
        raise SyntaxError(f"{str(e)}\n    - when casting output value to the output type")

    if isinstance(ret, ir.TupleExpr):
        vals = ret.val
    else:
        vals = [ret]

    if any(is_intf_id(v) for v in vals):
        return [ir.AssignValue(p, v) for p, v in zip(ctx.out_ports, vals)]

    stmts = []

    stmts.append(ir.Await('forward'))

    # Outputs values are offered in parallel. So first all outputs are declared
    # valid, and only then are all acknowledges awaited
    for p, v in zip(ctx.out_ports, vals):
        if isinstance(v, ir.ResExpr) and isinstance(v.val, ir.EmptyType):
            continue

        stmts.append(ir.AssignValue(ir.Component(p, 'data'), v))

    for p, v in zip(ctx.out_ports, vals):
        if isinstance(v, ir.ResExpr) and isinstance(v.val, ir.EmptyType):
            continue

        stmts.append(ir.Await(ir.Component(p, 'ready')))

    stmts.append(ir.Await('back'))

    return stmts


@node_visitor(ast.withitem)
def withitem(node: ast.withitem, ctx: Context):
    intf = visit_ast(node.context_expr, ctx)
    targets = visit_ast(node.optional_vars, ctx)

    if isinstance(intf, ir.ConcatExpr):
        data = ir.ConcatExpr([ir.Component(i, 'data') for i in intf.operands])
    else:
        data = ir.Component(intf, 'data')

    ass_targets = assign_targets(ctx, targets, data)

    return intf, ass_targets


@node_visitor(ast.AsyncWith)
def asyncwith(node, ctx: Context):
    assigns = [visit_ast(i, ctx) for i in node.items]

    intfs = []
    for intf, _ in assigns:
        if isinstance(intf, ir.ConcatExpr):
            intfs.extend(intf.operands)
        else:
            intfs.append(intf)

    for i in intfs:
        extend_stmts(ctx.ir_parent_block.stmts, ir.Await(ir.Component(i, 'valid')))

    for _, targets in assigns:
        extend_stmts(ctx.ir_parent_block.stmts, targets)

    # stmts = []
    for stmt in node.body:
        res_stmt = visit_ast(stmt, ctx)
        extend_stmts(ctx.ir_parent_block.stmts, res_stmt)
        # extend_stmts(stmts, res_stmt)

    # stmts.append(ir.Await('back'))
    extend_stmts(ctx.ir_parent_block.stmts, ir.Await('back'))
    for i in intfs:
        extend_stmts(
            ctx.ir_parent_block.stmts,
            ir.AssignValue(ir.Component(i, 'ready'), ir.res_true),
        )
        # stmts.append(ir.AssignValue(ir.Component(i, 'ready'), ir.res_true))

    # return stmts


class AsyncForContext:
    def __init__(self, intf, ctx):
        self.intf = intf
        self.ctx = ctx

    def __enter__(self):
        eot_name = self.ctx.find_unique_name('_eot')
        data_name = self.ctx.find_unique_name('_data')

        intf_type = self.intf.dtype.dtype

        self.ctx.scope[eot_name] = ir.Variable(eot_name, intf_type.eot)
        self.ctx.scope[data_name] = ir.Variable(data_name, intf_type.data)

        eot_init = ir.AssignValue(
            self.ctx.ref(eot_name),
            ir.ResExpr(intf_type.eot(0)),
        )

        eot_test = ir.BinOpExpr((self.ctx.ref(eot_name), ir.ResExpr(intf_type.eot.max)),
                                ir.opc.NotEq)

        eot_load = ir.AssignValue(self.ctx.ref(eot_name),
                                  ir.SubscriptExpr(ir.Component(self.intf, 'data'), ir.ResExpr(-1)))

        valid_await = ir.Await(ir.Component(self.intf, 'valid'))
        data_load = ir.AssignValue(self.ctx.ref(data_name), ir.Component(self.intf, 'data'))

        eot_loop_stmt = ir.LoopBlock(test=eot_test, stmts=[valid_await, data_load, eot_load])

        self.ctx.new_closure(eot_loop_stmt)

        return [eot_init, eot_loop_stmt]

    def __exit__(self, exception_type, exception_value, traceback):
        loop = self.ctx.closures.pop()

        loop.stmts.append(ir.Await('back'))
        loop.stmts.append(ir.AssignValue(ir.Component(self.intf, 'ready'), ir.res_true))
        merge_cond_alias_map(loop, self.ctx)


@node_visitor(ast.AsyncFor)
def AsyncFor(node, ctx: Context):
    out_intf_ref = visit_ast(node.iter, ctx)
    targets = visit_ast(node.target, ctx)

    with AsyncForContext(out_intf_ref, ctx) as stmts:
        extend_stmts(ctx.ir_parent_block.stmts,
                     assign_targets(ctx, targets, ir.Component(out_intf_ref, 'data')))

        for stmt in node.body:
            res_stmt = visit_ast(stmt, ctx)
            extend_stmts(ctx.ir_parent_block.stmts, res_stmt)

        return stmts


@node_visitor(ast.Await)
def _(node: ast.Await, ctx: Context):
    if isinstance(node.value, ast.Call):
        res = visit_ast(node.value.func, ctx)
        from pygears.sim import clk
        if isinstance(res, ir.ResExpr):
            if res.val == clk:
                return [ir.Await('back'), ir.Await('forward')]
            elif res.val.__qualname__ == 'Intf.ready':
                intf = res.val.__self__
                for p in ctx.gear.out_ports:
                    if p.producer is intf:
                        break
                else:
                    breakpoint()

                return ir.Await(ir.Component(ctx.ref(p.basename), 'ready'))

    breakpoint()



# #########################     LICENSE     ############################ #

# Copyright (c) 2005-2018, Michele Simionato
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

#   Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#   Redistributions in bytecode form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
"""
Decorator module, see http://pypi.python.org/pypi/decorator
for the documentation.
"""

import re
import sys
import inspect
import itertools

DEF = re.compile(r'\s*def\s*([_\w][_\w\d]*)\s*\(')


# basic functionality
class FunctionMaker(object):
    """
    An object with the ability to create functions with a given signature.
    It has attributes name, doc, module, signature, defaults, dict and
    methods update and make.
    """

    # Atomic get-and-increment provided by the GIL
    _compile_count = itertools.count()

    # make pylint happy
    args = varargs = varkw = defaults = kwonlyargs = kwonlydefaults = ()

    def __init__(self,
                 func=None,
                 name=None,
                 signature=None,
                 defaults=None,
                 doc=None,
                 module=None,
                 funcdict=None,
                 extra_kwds=None):
        self.shortsignature = signature
        if func:
            # func can be a class or a callable, but not an instance method
            self.name = func.__name__
            if self.name == '<lambda>':  # small hack for lambda functions
                self.name = '_lambda_'
            self.doc = func.__doc__
            self.module = func.__module__
            if inspect.isfunction(func):
                argspec = inspect.getfullargspec(func)
                self.annotations = getattr(func, '__annotations__', {})
                for a in ('args', 'varargs', 'varkw', 'defaults', 'kwonlyargs',
                          'kwonlydefaults'):
                    setattr(self, a, getattr(argspec, a))
                for i, arg in enumerate(self.args):
                    setattr(self, 'arg%d' % i, arg)
                allargs = list(self.args)
                allshortargs = list(self.args)
                if self.varargs:
                    allargs.append('*' + self.varargs)
                    allshortargs.append('*' + self.varargs)
                elif self.kwonlyargs or extra_kwds:
                    allargs.append('*')  # single star syntax
                if extra_kwds:
                    for a in extra_kwds:
                        allargs.append('%s=None' % a)
                        allshortargs.append('%s=%s' % (a, a))
                    if not self.kwonlydefaults:
                        self.kwonlydefaults = {}
                    self.kwonlydefaults.update(extra_kwds)
                for a in self.kwonlyargs:
                    allargs.append('%s=None' % a)
                    allshortargs.append('%s=%s' % (a, a))
                if self.varkw:
                    allargs.append('**' + self.varkw)
                    allshortargs.append('**' + self.varkw)
                self.signature = ', '.join(allargs)
                self.shortsignature = ', '.join(allshortargs)
                self.dict = func.__dict__.copy()
        # func=None happens when decorating a caller
        if name:
            self.name = name
        if signature is not None:
            self.signature = signature
        if defaults:
            self.defaults = defaults
        if doc:
            self.doc = doc
        if module:
            self.module = module
        if funcdict:
            self.dict = funcdict
        # check existence required attributes
        assert hasattr(self, 'name')
        if not hasattr(self, 'signature'):
            raise TypeError('You are decorating a non function: %s' % func)

    def update(self, func, **kw):
        "Update the signature of func with the data in self"
        func.__name__ = self.name
        func.__doc__ = getattr(self, 'doc', None)
        func.__dict__ = getattr(self, 'dict', {})
        func.__defaults__ = self.defaults
        func.__kwdefaults__ = self.kwonlydefaults or None
        func.__annotations__ = getattr(self, 'annotations', None)
        try:
            frame = sys._getframe(3)
        except AttributeError:  # for IronPython and similar implementations
            callermodule = '?'
        else:
            callermodule = frame.f_globals.get('__name__', '?')
        func.__module__ = getattr(self, 'module', callermodule)
        func.__dict__.update(kw)

    def make(self, src_templ, evaldict=None, addsource=False, **attrs):
        "Make a new function from a given template and update the signature"
        src = src_templ % vars(self)  # expand name and signature
        evaldict = evaldict or {}
        mo = DEF.search(src)
        if mo is None:
            raise SyntaxError('not a valid function template\n%s' % src)
        name = mo.group(1)  # extract the function name
        names = set(
            [name] +
            [arg.strip(' *') for arg in self.shortsignature.split(',')])
        for n in names:
            if n in ('_func_', '_call_'):
                raise NameError('%s is overridden in\n%s' % (n, src))

        if not src.endswith('\n'):  # add a newline for old Pythons
            src += '\n'

        # Ensure each generated function has a unique filename for profilers
        # (such as cProfile) that depend on the tuple of (<filename>,
        # <definition line>, <function name>) being unique.
        filename = '<decorator-gen-%d>' % (next(self._compile_count), )
        try:
            code = compile(src, filename, 'single')
            exec(code, evaldict)
        except Exception:
            print('Error in generated code:', file=sys.stderr)
            print(src, file=sys.stderr)
            raise
        func = evaldict[name]
        if addsource:
            attrs['__source__'] = src
        self.update(func, **attrs)
        return func

    @classmethod
    def create(cls,
               obj,
               body,
               evaldict,
               defaults=None,
               doc=None,
               module=None,
               addsource=True,
               isasync=False,
               extra_kwds=None,
               annotations=None,
               **attrs):
        """
        Create a function from the strings name, signature and body.
        evaldict is the evaluation dictionary. If addsource is true an
        attribute __source__ is added to the result. The attributes attrs
        are added, if any.
        """
        if isinstance(obj, str):  # "name(signature)"
            name, rest = obj.strip().split('(', 1)
            signature = rest[:-1]  # strip a right parens
            func = None
        else:  # a function
            name = None
            signature = None
            func = obj
        self = cls(
            func,
            name,
            signature,
            defaults,
            doc,
            module,
            extra_kwds=extra_kwds)

        if annotations is not None:
            self.annotations = annotations

        ibody = '\n'.join('    ' + line for line in body.splitlines())
        caller = evaldict.get('_call_')  # when called from `decorate`
        if isasync or (caller and inspect.iscoroutinefunction(caller)):
            body = ('async def %(name)s(%(signature)s):\n' + ibody).replace(
                'return', 'return await')
        else:
            body = 'def %(name)s(%(signature)s):\n' + ibody
        return self.make(body, evaldict, addsource, **attrs)



from pygears import alternative, gear, module
from .ccat import ccat
from .cart import cart
from .fmap import fmap
from .permute import permuted_apply
from pygears.typing import Int, Integer, Queue, Tuple, typeof, code, Uint, Bool, cast
from pygears.util.utils import quiter
from .fmaps import queuemap
from .mux import mux

TCfg = Tuple[{
    'start': Integer['w_start'],
    'cnt': Integer['w_cnt'],
    'incr': Integer['w_incr']
}]


@gear
def qenumerate(*din: b'din_t'
               ) -> Queue[Tuple['din_t', Uint['bitw(len(din)-1)']]]:
    @gear
    def mux_wrap(sel):
        return mux(sel, *din) | Tuple

    return qrange(len(din)) | queuemap(f=mux_wrap)


def rng_out_type(cfg, cnt_steps):
    if cnt_steps:
        return cfg[0] + cfg[1] + cfg[2]

    return max(cfg[0], cfg[1])


def qrange_out_type(cfg):
    if typeof(cfg, Tuple):
        base = Int if any(c.signed for c in cfg) else Uint

        return cast(max(cfg[0], cfg[1]), base)

    return cfg


@gear(enablement=b'inclusive==False')
async def qrange(
        cfg: Tuple[{
            'start': Integer,
            'stop': Integer,
            'step': Integer
        }],
        *,
        inclusive=False,
) -> Queue['qrange_out_type(cfg)']:

    cnt: module().tout.data
    cur_cnt: cfg.dtype[0]
    last: Bool

    async with cfg as c:
        cnt = module().tout.data(c[0])
        last = False
        while not last:
            cur_cnt = cnt
            cnt += c[2]

            last = cnt >= c[1]
            yield cur_cnt, last


@alternative(qrange)
@gear(enablement=b'inclusive==False')
async def qrange_start_stop(
        cfg: Tuple[{
            'start': Integer,
            'stop': Integer
        }],
        *,
        inclusive=False,
) -> Queue['qrange_out_type(cfg)']:

    cnt: module().tout.data
    cur_cnt: cfg.dtype[0]
    last: Bool

    async with cfg as c:
        cnt = module().tout.data(c[0])
        last = False
        while not last:
            cur_cnt = cnt
            cnt += 1

            last = cnt == c[1]
            yield cur_cnt, cnt == c[1]


@alternative(qrange)
@gear(enablement=b'inclusive==True')
async def qrange_start_stop_inclusive(
        cfg: Tuple[{
            'start': Integer,
            'stop': Integer
        }],
        *,
        inclusive,
) -> Queue['qrange_out_type(cfg)']:
    cnt: module().tout.data
    last: Bool

    async with cfg as c:
        last = False
        cnt = module().tout.data(c[0])
        while not last:
            last = cnt == c[1]
            yield cnt, last
            cnt += 1


@alternative(qrange)
@gear(enablement=b'inclusive==False')
async def qrange_stop(stop: Integer, *, inclusive=False) -> Queue[b'stop']:
    cnt = stop.dtype(0)
    cur_cnt: stop.dtype
    last: Bool

    async with stop as s:
        last = False
        while not last:
            # while (cnt != s):
            cur_cnt = cnt
            cnt += 1

            last = cnt == s
            yield cur_cnt, last


@alternative(qrange)
@gear(enablement=b'inclusive==True')
async def qrange_stop_inclusive(stop: Integer, *, inclusive) -> Queue[b'stop']:
    cnt = stop.dtype(0)
    last: Bool

    async with stop as s:
        last = False
        while not last:
            last = cnt == s
            yield cnt, last
            cnt = code(cnt + 1, stop.dtype)



from pygears.typing import Fixp, Fixpnumber, Integer, Tuple, Ufixp, Uint, typeof, Int, Union, cast, Array, Queue, Maybe, Bool
from . import ir
from pygears.typing import get_match_conds, TypeMatchError


def fixp_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    val_dtype = opexp.dtype

    if typeof(val_dtype, Integer):
        other_cls = Fixp if val_dtype.signed else Ufixp
        val_dtype = other_cls[val_dtype.width, val_dtype.width]

    if cast_to is Fixpnumber:
        return ir.CastExpr(opexp, val_dtype)

    val_fract = val_dtype.fract
    fract = cast_to.fract

    if val_dtype.signed:
        opexp = ir.CastExpr(opexp, Int[val_dtype.width])
    else:
        opexp = ir.CastExpr(opexp, Uint[val_dtype.width])

    if fract > val_fract:
        shift = ir.BinOpExpr([opexp, ir.ResExpr(Uint(fract - val_fract))], ir.opc.LShift)
    else:
        shift = ir.BinOpExpr([opexp, ir.ResExpr(Uint(val_fract - fract))], ir.opc.RShift)

    return ir.CastExpr(shift, cast_to)


def subscript(opexp, index):
    if not isinstance(index, slice) and isinstance(opexp, ir.ConcatExpr):
        return opexp.operands[index]
    else:
        return ir.SubscriptExpr(opexp, ir.ResExpr(index))


def tuple_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    cast_fields = []
    for i in range(len(opexp.dtype)):
        field = subscript(opexp, i)
        cast_fields.append(resolve_cast_func(field, cast_to[i]))

    return ir.ConcatExpr(cast_fields)


def queue_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    if cast_to == opexp.dtype:
        return opexp

    return ir.CastExpr(
        ir.ConcatExpr([resolve_cast_func(subscript(opexp, 0), cast_to.data), subscript(opexp, 1)]),
        cast_to)


def array_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    dlen, clen = len(opexp.dtype), len(cast_to)

    if dlen == clen:
        cast_fields = []
        for i in range(len(opexp.dtype)):
            field = subscript(opexp, i)
            cast_fields.append(resolve_cast_func(field, cast_to.data))
    else:
        elen = len(cast_to.data)
        cast_fields = []
        for i in range(len(cast_to)):
            field = subscript(opexp, slice(elen*i, elen*(i+1)))
            cast_fields.append(resolve_cast_func(field, cast_to.data))

    return ir.ConcatExpr(cast_fields)


def union_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    op_data = subscript(opexp, 0)
    op_ctrl = subscript(opexp, 1)

    cast_fields = []
    uint_data = ir.CastExpr(op_data, Uint[op_data.dtype.width])
    cast_fields.append(resolve_cast_func(uint_data, cast_to.data))

    cast_fields.append(resolve_cast_func(op_ctrl, cast_to.ctrl))

    return ir.CastExpr(ir.ConcatExpr(cast_fields), cast_to)


def uint_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    if typeof(opexp.dtype, Ufixp):
        if opexp.dtype.fract >= 0:
            opexp = ir.BinOpExpr((opexp, ir.ResExpr(opexp.dtype.fract)), ir.opc.RShift)
        else:
            opexp = ir.BinOpExpr((opexp, ir.ResExpr(-opexp.dtype.fract)), ir.opc.LShift)

    return ir.CastExpr(opexp, cast_to)


def int_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)
    if typeof(opexp.dtype, Uint):
        opexp = ir.CastExpr(opexp, Uint[cast_to.width])

    return ir.CastExpr(opexp, cast_to)


# # TODO: Generalize this to any Union
# def maybe_resolver(opexp, cast_to):
#     if cast_to.specified and opexp.dtype == cast_to.dtype:
#         data = ir.CastExpr(resolve_cast_func(opexp, cast_to.dtype), cast_to[0])
#         return ir.CastExpr(ir.ConcatExpr([data, ir.ResExpr(Bool(True))]), cast_to)

#     cast_to = cast(opexp.dtype, cast_to)

#     if opexp.dtype == cast_to.dtype:
#         data = ir.CastExpr(resolve_cast_func(opexp, cast_to.dtype), cast_to[0])
#         return ir.CastExpr(ir.ConcatExpr([data, ir.ResExpr(Bool(True))]), cast_to)
#     elif typeof(opexp.dtype, Tuple):
#         data = resolve_cast_func(ir.SubscriptExpr(opexp, ir.ResExpr(0)), cast_to.dtype)
#         ctrl = resolve_cast_func(ir.SubscriptExpr(opexp, ir.ResExpr(1)), cast_to[1])
#         return ir.CastExpr(ir.ConcatExpr([data, ctrl]), cast_to)
#     elif typeof(opexp.dtype, Union):
#         return ir.CastExpr(opexp, cast_to)
#     else:
#         breakpoint()


def maybe_resolver(opexp, cast_to):
    cast_to = cast(opexp.dtype, cast_to)

    if typeof(opexp.dtype, Tuple):
        data = resolve_cast_func(ir.SubscriptExpr(opexp, ir.ResExpr(0)), cast_to.dtype)
        ctrl = resolve_cast_func(ir.SubscriptExpr(opexp, ir.ResExpr(1)), cast_to[1])
        return ir.CastExpr(ir.ConcatExpr([data, ctrl]), cast_to)
    elif typeof(opexp.dtype, Union):
        return ir.CastExpr(opexp, cast_to)
    else:
        breakpoint()

resolvers = {
    Fixpnumber: fixp_resolver,
    Tuple: tuple_resolver,
    Queue: queue_resolver,
    Array: array_resolver,
    Int: int_resolver,
    Uint: uint_resolver,
    Maybe: maybe_resolver, # Has to go before Union to be detected!
    Union: union_resolver,
}


def resolve_cast_func(opexp, dtype):
    if opexp.dtype == dtype:
        return opexp

    for templ in resolvers:
        try:
            get_match_conds(dtype, templ)
            return resolvers[templ](opexp, dtype)
        except TypeMatchError:
            continue

    return ir.CastExpr(operand=opexp, cast_to=cast(opexp.dtype, dtype))



import inspect
import fnmatch
import sys
from copy import copy

from pygears.core.graph import get_producer_port
from pygears.conf import MultiAlternativeError, core_log, reg
from pygears.typing import Any, cast, get_match_conds
from pygears.core.util import is_standard_func, get_function_context_dict

from .partial import Partial
from .intf import Intf
from .infer_ftypes import TypeMatchError, infer_ftypes, type_is_specified
from .gear import TooManyArguments, GearTypeNotSpecified, GearArgsNotSpecified
from .gear import Gear, create_hier
from .gear_decorator import GearDecoratorPlugin
from .gear_memoize import get_memoized_gear, memoize_gear
from .port import HDLConsumer, HDLProducer
from .channel import channel_interfaces


def check(pattern, node):
    if isinstance(pattern, str):
        return fnmatch.fnmatch(node.name, pattern)
    else:
        return pattern(node)

def is_traced(node):
    return any(check(p, node) for p in reg['debug/trace'])

    # cfg = reg['debug/trace']
    # return any(fnmatch.fnmatch(name, p) for p in cfg)


def get_obj_var_name(frame, obj):
    for var_name, var_obj in frame.f_locals.items():
        if obj is var_obj:
            return var_name
    else:
        None


def find_current_gear_frame():
    import inspect
    code_map = reg['gear/code_map']
    if not code_map:
        return None

    for frame, *_ in inspect.stack():
        if frame.f_code is code_map[-1].func.__code__:
            return frame
    else:
        return None


def check_args_num(argnames, varargsname, args):
    if (len(args) < len(argnames)) or (not varargsname and (len(args) > len(argnames))):
        balance = "few" if (len(args) < len(argnames)) else "many"

        raise TooManyArguments(f"Too {balance} arguments provided.")


def check_args_specified(args):
    for name, intf in args.items():
        if not isinstance(intf, Intf):
            raise GearArgsNotSpecified(f'Unresolved input argument "{name}": {repr(intf)}')

        if not type_is_specified(intf.dtype):
            raise GearArgsNotSpecified(
                f'Input argument "{name}" has unresolved type "{repr(intf.dtype)}"')


def resolve_gear_name(func, __base__):
    if __base__ is None:
        name = func.__name__
    else:
        name = __base__.__name__

    return name


oper_name = {
    '__add__': '+',
    '__and__': '&',
    '__floordiv__': '/',
    '__truediv__': '//',
    '__eq__': '==',
    '__ge__': '>=',
    '__gt__': '>',
    '__invert__': '~',
    '__le__': '<=',
    '__lt__': '<',
    '__lshift__': '<<',
    '__matmul__': '@',
    '__mod__': '%',
    '__mul__': '*',
    '__ne__': '!=',
    '__neg__': '-',
    '__rshift__': '>>',
    '__sub__': '-',
    '__xor__': '^',
}


def get_operator_name(func):
    for name, p in reg['gear/intf_oper'].items():
        if isinstance(p, Partial):
            p = p.func

        if p is func:
            return oper_name[name]

    return None


def infer_const_args(args):
    args_res = {}
    const_args = {}
    for name, intf in args.items():
        if not isinstance(intf, Intf):
            from pygears.lib.const import get_literal_type
            try:
                const_args[name] = intf
                intf = Intf(get_literal_type(intf))
            except GearTypeNotSpecified:
                if isinstance(intf, Partial):
                    # raise GearArgsNotSpecified(
                    #     f'Unresolved gear "{intf.func.__name__}" with inputs'
                    #     f' {intf.args} and parameters {intf.kwds},'
                    #     f'connected to the input "{name}"')

                    opinfo = ''
                    op = get_operator_name(intf.func)
                    if op:
                        opinfo = f' (operator "{op}")'

                    raise GearArgsNotSpecified(
                        f'Unresolved gear "{intf.func.__name__}"{opinfo} with'
                        f' arguments {intf.args} and parameters {intf.kwds},'
                        f' connected to the input "{name}": {str(MultiAlternativeError(intf.errors))}'
                    )
                else:
                    raise GearArgsNotSpecified(
                        f'Unresolved argument "{intf}" connected to the input'
                        f' "{name}"')

        args_res[name] = intf

    return args_res, const_args


def expand_varargs(args, annotations, varargsname, varargs):
    vararg_type_list = []
    if varargsname in annotations:
        vararg_type = annotations[varargsname]
    else:
        vararg_type = Any

    # Append the types of the varargsname
    for i, a in enumerate(varargs):
        if isinstance(vararg_type, str):
            # If vararg_type is a template string, it can be made
            # dependent on the arguments position
            type_tmpl_i = vararg_type.format(i).encode()
        else:
            # Vararg is not a template and should be passed as is
            type_tmpl_i = vararg_type

        argname = f'{varargsname}{i}'

        vararg_type_list.append(argname)
        annotations[argname] = type_tmpl_i
        args[argname] = a

    if vararg_type_list:
        annotations[varargsname] = f'({", ".join(vararg_type_list)}, )'.encode()


def resolve_return_annotation(annotations):
    if "return" in annotations:
        ret_anot = annotations["return"]
        if isinstance(ret_anot, dict):
            annotations['return'] = tuple(ret_anot.values())
            if len(annotations['return']) == 1:
                annotations['return'] = annotations['return'][0]
    else:
        annotations['return'] = None


def resolve_args(args, argnames, annotations, varargs):
    check_args_num(argnames, varargs, args)
    args_dict = {name: a for name, a in zip(argnames, args)}

    if varargs:
        expand_varargs(args_dict, annotations, varargs, args[len(args_dict):])

    resolve_return_annotation(annotations)

    for a in args_dict:
        if a not in annotations:
            annotations[a] = Any

    return args_dict, annotations


def gear_signature(func, args, kwds):
    paramspec = inspect.getfullargspec(func)

    args, annotations = resolve_args(args, paramspec.args, paramspec.annotations, paramspec.varargs)

    kwddefaults = paramspec.kwonlydefaults or {}

    templates = {**annotations, **kwddefaults, **kwds}

    return args, templates


def infer_params(args, params, context):
    arg_types = {name: arg.dtype for name, arg in args.items()}

    return infer_ftypes(params, arg_types, namespace=context)


class intf_name_tracer:
    def __init__(self, gear):
        self.enabled = reg['gear/infer_signal_names']
        if self.enabled == 'debug':
            self.enabled = is_traced(gear)

        if not self.enabled:
            return

        self.code_map = reg['gear/code_map']
        self.gear = gear

    def tracer(self, frame, event, arg):
        if event == 'return':
            for cm in self.code_map:
                if frame.f_code is cm.func.__code__:
                    cm.func_locals = frame.f_locals.copy()

    def __enter__(self):
        if not self.enabled:
            return

        self.code_map.append(self.gear)

        # tracer is activated on next call, return or exception
        if reg['gear/current_module'].parent == reg['gear/root']:
            sys.setprofile(self.tracer)

        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if not self.enabled:
            return

        if reg['gear/current_module'].parent == reg['gear/root']:
            sys.setprofile(None)

        cm = self.code_map.pop()

        if exception_type is None and hasattr(cm, 'func_locals'):
            for name, val in filter(lambda x: isinstance(x[1], Intf), cm.func_locals.items()):
                if not hasattr(val, 'var_name'):
                    val.var_name = name


# TODO: Apparently no error is thrown if an input interface is not connected
# within a hierarchical module
def resolve_func(gear_inst):
    out_dtype = gear_inst.params.get('return', None)
    if out_dtype is None:
        out_dtype = ()
    elif out_dtype:
        if isinstance(out_dtype, dict):
            out_dtype = tuple(out_dtype.values())

    if not isinstance(out_dtype, tuple):
        out_dtype = (out_dtype, )

    if not is_standard_func(gear_inst.func):
        # TODO: If something is being yield-ed and there is no output type
        # specified, try to throw an error
        return tuple(), out_dtype

    with create_hier(gear_inst):
        # TODO: Try to detect infinite recursions
        # TODO: If the gear is instantiated in REPL, intf_name_tracer will fail
        with intf_name_tracer(gear_inst):
            local_in_intfs = gear_inst.in_port_intfs
            out_intfs = gear_inst.func(*local_in_intfs, **gear_inst.explicit_params)

        if out_intfs is None:
            out_intfs = tuple()
        elif not isinstance(out_intfs, tuple):
            out_intfs = (out_intfs, )

        for i, intf in enumerate(out_intfs):
            if isinstance(intf, Partial):
                raise GearArgsNotSpecified(
                    f'Unresolved gear "{intf.func.__name__}" with'
                    f' arguments {intf.args} and parameters {intf.kwds},'
                    f' returned as output "{i}": {str(MultiAlternativeError(intf.errors))}')

        err = None
        try:
            out_intfs, out_dtype = resolve_out_types(out_intfs, out_dtype, gear_inst)
        except (TypeError, TypeMatchError) as e:
            err = type(e)(f"{str(e)}\n    when instantiating '{gear_inst.name}'")

        if err:
            raise err

    return out_intfs, out_dtype


def resolve_gear(gear_inst, out_intfs, out_dtype, fix_intfs):
    dflt_dout_name = reg['gear/naming/default_out_name']
    for i in range(len(gear_inst.outnames), len(out_dtype)):
        if out_intfs and hasattr(out_intfs[i], 'var_name'):
            gear_inst.outnames.append(out_intfs[i].var_name)
        else:
            gear_inst.outnames.append(dflt_dout_name if len(out_dtype) ==
                                      1 else f'{dflt_dout_name}{i}')

    gear_inst.connect_output(out_intfs, out_dtype)

    # Connect output interfaces
    intfs = []
    out_intfs = []
    if isinstance(fix_intfs, dict):
        for i, (name, dt) in enumerate(zip(gear_inst.outnames, out_dtype)):
            if (name in fix_intfs) or (i in fix_intfs):
                if name in fix_intfs:
                    intf = fix_intfs[name]
                elif i in fix_intfs:
                    intf = fix_intfs[i]

                err = None
                try:
                    get_match_conds(dt, intf.dtype)
                except (TypeError, TypeMatchError) as e:
                    err = type(
                        e
                    )(f"{str(e)}\n    when connecting user supplied output interface '{name}' of '{gear_inst.name}'"
                      f"\n    FIX: Consider changing the type of the supplied output interface '{name}' to '{repr(dt)}'"
                      )

                if err:
                    raise err
            else:
                intf = Intf(dt)
                out_intfs.append(intf)

            intfs.append(intf)

    elif fix_intfs:
        # TODO: Should we allow partially supplied fix_intfs? Maybe None should
        # be supplied where a new Intf should be created
        # TODO: Do similar type checking as above
        intfs = fix_intfs
    else:
        intfs = [Intf(dt) for dt in out_dtype]
        out_intfs = intfs

    if len(intfs) != len(gear_inst.out_port_intfs):
        raise GearArgsNotSpecified(
            f'Received {len(intfs)} output interfaces,'
            f' while expecting {len(gear_inst.out_port_intfs)}'
            f"\n    when instantiating '{gear_inst.name}'"
        )

    for intf, port in zip(intfs, gear_inst.out_ports):
        intf.source(port)

    if any(not type_is_specified(i.dtype) for i in out_intfs):
        raise GearTypeNotSpecified(f'Output type of the gear "{gear_inst.name}"'
                                   f' could not be resolved, and resulted in "{repr(out_dtype)}"')

    for c in gear_inst.child:
        channel_interfaces(c)

    if len(out_intfs) > 1:
        return tuple(out_intfs)
    elif len(out_intfs) == 1:
        return out_intfs[0]
    else:
        return None


def resolve_out_types(out_intfs, out_dtype, gear_inst):

    if out_intfs and (not out_dtype):
        out_dtype = tuple(intf.dtype for intf in out_intfs)
        return out_intfs, out_dtype

    if out_intfs:
        if len(out_intfs) != len(out_dtype):
            relation = 'smaller' if len(out_intfs) < len(out_dtype) else 'larger'
            raise TypeMatchError(
                f"Number of actual output interfaces ({len(out_intfs)}) is {relation} "
                f"than the number of specified output types: ({tuple(getattr(i, 'dtype', type(i)) for i in out_intfs)})"
                f" vs {repr(out_dtype)}")

        casted_out_intfs = list(out_intfs)

        # Try casting interface types upfront to get better error messaging
        for i, (intf, t) in enumerate(zip(out_intfs, out_dtype)):
            err = None
            try:
                if intf.dtype != t:
                    cast(intf.dtype, t)
            except (TypeError, TypeMatchError) as e:
                err = type(e)(f"{str(e)}, when casting type for output port {i}")

            if err:
                raise err

        # If no exceptions occured, do it for real
        for i, (intf, t) in enumerate(zip(out_intfs, out_dtype)):
            if intf.dtype != t:
                from pygears.lib.cast import cast as cast_gear
                casted_out_intfs[i] = cast_gear(intf, t=t)

        out_intfs = tuple(casted_out_intfs)
        return out_intfs, out_dtype

    return out_intfs, out_dtype


def terminate_internal_intfs(gear_inst):
    if not is_standard_func(gear_inst.func):
        for p in gear_inst.in_ports:
            if p.consumer is not None:
                p.consumer.connect(HDLConsumer())

        for i in gear_inst.out_port_intfs:
            i.source(HDLProducer())


def gear_base_resolver(func, *args, name=None, intfs=None, **kwds):
    meta_kwds = func.meta_kwds
    name = name or resolve_gear_name(func, meta_kwds['__base__'])

    err = None
    try:
        args, param_templates = gear_signature(func, args, kwds)

        param_templates['_enablement'] = meta_kwds['enablement']
        args, const_args = infer_const_args(args)
        check_args_specified(args)
    except (TooManyArguments, GearArgsNotSpecified) as e:
        err = type(e)(f'{str(e)}\n - when instantiating "{name}"')

    if err:
        raise err

    if intfs is None:
        fix_intfs = []
    elif isinstance(intfs, Intf):
        fix_intfs = [intfs]
    elif isinstance(intfs, dict):
        fix_intfs = intfs
    else:
        fix_intfs = list(intfs)

    if reg['gear/memoize']:
        gear_inst, outputs, memo_key = get_memoized_gear(func, args, const_args, kwds, fix_intfs,
                                                         name)

        if gear_inst is not None:
            if len(outputs) == 1:
                return outputs[0]
            else:
                return outputs

    try:
        params = infer_params(args, param_templates, context=get_function_context_dict(func))
    except TypeMatchError as e:
        err = type(e)(f'{str(e)}\n - when instantiating "{name}"')

    if err:
        raise err

    if not err:
        if not params.pop('_enablement'):
            err = TypeMatchError(f'Enablement condition failed for "{name}" alternative'
                                 f' "{meta_kwds["definition"].__module__}.'
                                 f'{meta_kwds["definition"].__name__}": '
                                 f'{meta_kwds["enablement"].decode()}')

    params['name'] = name
    params['intfs'] = fix_intfs

    gear_inst = Gear(func, params)

    if err:
        err.gear = gear_inst
        err.root_gear = gear_inst

    if not err:
        gear_inst.connect_input(args, const_args)
        try:
            out_intfs, out_dtype = resolve_func(gear_inst)
            out_intfs = resolve_gear(gear_inst, out_intfs, out_dtype, fix_intfs)
            terminate_internal_intfs(gear_inst)

        except (TooManyArguments, GearTypeNotSpecified, GearArgsNotSpecified, TypeError,
                TypeMatchError, MultiAlternativeError) as e:
            err = e
            if hasattr(func, 'alternatives') or hasattr(func, 'alternative_to'):
                err.root_gear = gear_inst
        except Exception as e:
            if not hasattr(e, '_stamped'):
                e.args = (f'{str(e)}, in the module "{reg["gear/current_module"].name}/{name}"', )
                e._stamped = True

            err = e

    if err:
        if hasattr(func, 'alternatives') or hasattr(func, 'alternative_to'):
            gear_inst.parent.child.remove(gear_inst)
            for port in gear_inst.in_ports:
                if port.basename not in gear_inst.const_args:
                    port.producer.consumers.remove(port)
                else:
                    gear_inst.parent.child.remove(get_producer_port(port).gear)

        raise err

    if reg['gear/memoize'] and not func.__name__.endswith('_unpack__'):
        if memo_key is not None:
            memoize_gear(gear_inst, memo_key)

    return out_intfs


class GearInstPlugin(GearDecoratorPlugin):
    @classmethod
    def bind(cls):
        reg['gear/code_map'] = []
        reg['gear/gear_dflt_resolver'] = gear_base_resolver
        reg.confdef('gear/memoize', False)
        reg.confdef('gear/infer_signal_names', 'debug')
        reg.confdef('debug/trace', default=[])

    @classmethod
    def reset(cls):
        reg['gear/code_map'] = []
